#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# PatchBay Canvas engine using QGraphicsView/Scene
# Copyright (C) 2019-2022 Mathieu Picot <picotmathieu@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of
# the License, or any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# For a full copy of the GNU General Public License see the doc/GPL.txt file.

from enum import Enum
from typing import Union
from PyQt5.QtCore import pyqtSlot, QCoreApplication
from PyQt5.QtWidgets import QWidgetAction, QMenu, QAction
from PyQt5.QtGui import QIcon, QPixmap, QMouseEvent, QColor

from .utils import (get_group_icon, get_portgroup_short_name_splitted,
                    get_portgroup_connection_state, connect_portgroups,
                    connect_ports, canvas_callback, get_portgroup_full_name,
                    get_full_port_name, is_dark_theme, get_portgroup_port_list)
from .init_values import (
    GroupObject,
    BoxType,
    PortObject,
    PortSubType,
    PortgrpObject,
    canvas,
    ClipboardElement,
    CallbackAct,
    PortType,
    PortMode,
    ConnectionObject)
from .connect_menu_widgets import CheckFrame


_translate = QCoreApplication.translate


class _Dangerous(Enum):
    NO_CARE = 0
    NO = 1
    YES = 2


class _DataConnElement:
    port_id: int
    portgrp_id: int
    action: QWidgetAction
    check_frame: CheckFrame


class _DataDisconnElement:
    group_id: int
    portgrp_id: int
    port_id_list: list[int]
    action: QAction


class SubMenu(QMenu):
    def __init__(self, name: str, p_object: Union[PortObject, PortgrpObject], parent):
        QMenu.__init__(self, name, parent)
        self._p_object = p_object

    def connection_asked_from_box(
            self, p_object: Union[PortObject, PortgrpObject], connect: bool):
        pass


class GroupConnectMenu(SubMenu):
    def __init__(self, group: GroupObject, p_object: Union[PortObject, PortgrpObject],
                 parent: 'SubMenu',
                 dangerous_mode=_Dangerous.NO_CARE):
        short_group_name = group.group_name
        
        if len(short_group_name) > 15 and '/' in short_group_name:
            short_group_name = short_group_name.partition('/')[2]

        SubMenu.__init__(self, short_group_name, p_object, parent)
        
        po = self._p_object
        
        self.setIcon(
            get_group_icon(group.group_id, po.port_mode.opposite(), is_dark_theme(self)))
        self.hovered.connect(self._mouse_hover_menu)
        
        self._parent = parent
        self._group = group
        self._elements = list[_DataConnElement]()
        self._dangerous_mode = dangerous_mode

        self._last_portgrp_id = 0
        
        theme = canvas.theme.box
        if group.box_type == BoxType.CLIENT:
            theme = theme.client
        elif group.box_type == BoxType.HARDWARE:
            theme = theme.hardware

        bg_color = theme.background_color().name(QColor.HexArgb)
        border_color = theme.fill_pen().color().name(QColor.HexArgb)
        
        self.setStyleSheet(
            f"QMenu{{background-color:{bg_color}; border: 2px solid {border_color}; border-radius:4px}}")
        self.setMinimumHeight(50)
        self._elements_added = False

    def group_id(self) -> int:
        return self._group.group_id

    def add_all_elements(self):
        if self._elements_added:
            return

        po = self._p_object
        
        for port in canvas.list_ports(group_id=self._group.group_id):
            if (port.port_type is po.port_type
                    and port.port_mode is not po.port_mode):
                if isinstance(po, PortgrpObject) and port.portgrp_id:
                    if port.portgrp_id != self._last_portgrp_id:
                        for portgrp in canvas.list_portgroups(group_id=port.group_id):
                            if portgrp.portgrp_id == port.portgrp_id:
                                pg_name, pts_name = get_portgroup_short_name_splitted(
                                    portgrp.group_id, portgrp.portgrp_id)
                                
                                self._add_element(portgrp, pg_name, pts_name)
                                break
                else:
                    if (self._dangerous_mode is _Dangerous.YES
                            and po.port_subtype == port.port_subtype):
                        continue

                    if (self._dangerous_mode is _Dangerous.NO
                            and po.port_subtype != port.port_subtype):
                        continue

                    self._add_element(port, port.port_name, '')
        
        self._elements_added = True

    def _add_element(self, p_object: Union[PortObject, PortgrpObject],
                    port_name: str, port_name_end: str):
        if p_object.port_subtype is PortSubType.CV:
            port_name = f"CV | {port_name}"

        check_frame = CheckFrame(p_object, port_name, port_name_end, self)
        action = QWidgetAction(self)
        action.setDefaultWidget(check_frame)
        self.addAction(action)

        self._last_portgrp_id = p_object.portgrp_id

        check_frame.set_check_state(
            get_portgroup_connection_state(
                self._p_object.group_id, self._p_object.get_port_ids(),
                p_object.group_id, p_object.get_port_ids()))

        element = _DataConnElement()
        element.port_id = p_object.port_id if isinstance(p_object, PortObject) else -1
        element.portgrp_id = p_object.portgrp_id
        element.action = action
        element.check_frame = check_frame
        self._elements.append(element)

    def _remove_element(self, port_id: int, portgrp_id: int):
        for element in self._elements:
            if (element.port_id == port_id
                    and element.portgrp_id == portgrp_id):
                self.removeAction(element.action)
                self._elements.remove(element)
                break

    def count_elements(self) -> int:
        return len(self._elements)

    def get_first_element(self) -> _DataConnElement:
        return self._elements[0]

    def check_element(self, port_id: int, portgrp_id: int, check_state: int):
        for element in self._elements:
            if (element.port_id == port_id
                    and element.portgrp_id == portgrp_id):
                element.check_frame.set_check_state(check_state)
                break

    def connection_asked_from_box(self, p_object: Union[PortObject, PortgrpObject],
                                  yesno: bool):
        self._parent.connection_asked_from_box(p_object, yesno)

    def keyPressEvent(self, event) -> None:
        return super().keyPressEvent(event)

    def _mouse_hover_menu(self, action: QWidgetAction):
        action.defaultWidget().setFocus()


class DangerousMenu(SubMenu):
    def __init__(self, name, p_object: Union[PortObject, PortgrpObject], parent):
        SubMenu.__init__(self, name, p_object, parent)
        self.setIcon(QIcon.fromTheme('emblem-warning'))

        self.group_menus = list[GroupConnectMenu]()
        self.connection_list = list[ConnectionObject]()

    def add_group_menu(self, group: GroupObject):
        group_menu = GroupConnectMenu(group, self._p_object, self,
                                      dangerous_mode=_Dangerous.YES)
        self.group_menus.append(group_menu)
        self.addMenu(group_menu)

    def connection_asked_from_box(self, p_object: Union[PortObject, PortgrpObject],
                                  yesno: bool):
        po = self._p_object
        
        if yesno:
            if isinstance(po, PortgrpObject):
                # in and out are portgroups
                connect_portgroups(po, p_object)
            else:
                for self_port_id in po.get_port_ids():
                    connect_ports(po.group_id, self_port_id,
                                        p_object.group_id, p_object.port_id)
        else:
            if isinstance(po, PortgrpObject):
                connect_portgroups(po, p_object, disconnect=True)
            else:
                # port = canvas.get_port(po.group_id, port_id)
                # if port:
                    for connection in canvas.list_connections(po, p_object):
                        canvas_callback(
                            CallbackAct.PORTS_DISCONNECT,
                            connection.connection_id)
    
    def mouseMoveEvent(self, event: QMouseEvent):
        action = self.actionAt(event.pos())
        if action:
            menu = action.menu()
            if isinstance(menu, GroupConnectMenu):
                menu.add_all_elements()
        
        return super().mouseMoveEvent(event)
    

class ConnectMenu(SubMenu):
    def __init__(self, p_object: Union[PortObject, PortgrpObject], parent=None):
        SubMenu.__init__(self, _translate('patchbay', 'Connect'),
                         p_object, parent)

        self.group_menus = list[GroupConnectMenu]()
        self.connection_list = list[ConnectionObject]()

        self.hovered.connect(self._hovered)

        dangerous_name = ''
        has_dangerous_global = False

        po = p_object

        if po.port_type is PortType.AUDIO_JACK:
            if (po.port_mode is PortMode.OUTPUT
                    and po.port_subtype is PortSubType.CV):
                dangerous_name = _translate(
                    'patchbay', 'Audio | DANGEROUS !!!')
            elif (po.port_mode is PortMode.INPUT
                    and po.port_subtype is PortSubType.REGULAR):
                dangerous_name = _translate(
                    'patchbay', 'CV | DANGEROUS !!!')

        self.dangerous_submenu = DangerousMenu(
            dangerous_name, self._p_object, self)

        # add the needed groups (not the ports)
        for group in canvas.group_list:
            grp_has_dangerous = False
            grp_has_regular = False

            for port in canvas.list_ports(group_id=group.group_id):
                if (port.port_type is po.port_type
                        and port.port_mode is not po.port_mode):

                    if (po.port_type is PortType.AUDIO_JACK
                            and ((po.port_mode is PortMode.OUTPUT
                                  and po.port_subtype is PortSubType.CV
                                  and port.port_subtype is PortSubType.REGULAR)
                                 or (po.port_mode is PortMode.INPUT
                                     and not po.port_subtype is PortSubType.CV
                                     and port.port_subtype is PortSubType.CV))):
                        if not grp_has_dangerous:
                            self.dangerous_submenu.add_group_menu(group)
                        grp_has_dangerous = True
                        has_dangerous_global = True
                    else:
                        if not grp_has_regular:
                            self.add_group_menu(group)
                        grp_has_regular = True

                    if grp_has_dangerous and grp_has_regular:
                        break
                
        if has_dangerous_global:
            self.addSeparator()
            self.addMenu(self.dangerous_submenu)
            
        self._elements_width = 100

    def add_group_menu(self, group: GroupObject):
        po = self._p_object
        dangerous = _Dangerous.NO_CARE
        if (po.port_type == PortType.AUDIO_JACK
                and ((po.port_mode is PortMode.OUTPUT
                      and po.port_subtype is PortSubType.CV)
                     or (po.port_mode is PortMode.INPUT
                         and not po.port_subtype is PortSubType.CV))):
            dangerous = _Dangerous.NO

        group_menu = GroupConnectMenu(group, self._p_object, self,
                                      dangerous_mode=dangerous)
        self.group_menus.append(group_menu)
        self.addMenu(group_menu)

    def connection_asked_from_box(self, p_object: Union[PortObject, PortgrpObject], yesno: bool):
        po = self._p_object
        
        if yesno:
            if isinstance(po, PortgrpObject):
                connect_portgroups(po, p_object)
            else:
                for self_port_id in po.get_port_ids():
                    connect_ports(po.group_id, self_port_id,
                                        p_object.group_id, p_object.port_id)
        else:
            if isinstance(po, PortgrpObject):
                connect_portgroups(po, p_object, disconnect=True)
            else:
                for connection in canvas.list_connections(po, p_object):
                    canvas_callback(
                        CallbackAct.PORTS_DISCONNECT,
                        connection.connection_id)

    def leaveEvent(self, event):
        # prevent to close the menu accidentaly when the mouse 
        # leaves the menu area
        pass
    
    def mouseMoveEvent(self, event: QMouseEvent):
        action = self.actionAt(event.pos())
        if action:
            menu = action.menu()
            if isinstance(menu, GroupConnectMenu):
                menu.add_all_elements()
        
        return super().mouseMoveEvent(event)

    def _hovered(self, action: QAction):
        menu = action.menu()
        if isinstance(menu, GroupConnectMenu):
            menu.setMinimumWidth(self._elements_width)
            menu.add_all_elements()
            self._elements_width = max(self._elements_width, menu.width())

    # TODO was initially added the fact menu was updated
    # when port was added or removed
    # for portgroup it seems to be much complicated
    # user will just have to re-open the menu


class DisconnectMenu(SubMenu):
    def __init__(self, p_object: Union[PortObject, PortgrpObject], parent):
        SubMenu.__init__(self, _translate('patchbay', "Disconnect"),
                         p_object, parent)
        self._elements = list[_DataDisconnElement]()

        self._no_action_title = _translate('patchbay', 'No connections')
        self._no_action = self.addAction(self._no_action_title)
        self._no_action.setEnabled(False)

    @pyqtSlot()
    def apply_disconnection(self):
        action = self.sender()
        if action is None:
            return

        for element in self._elements:
            if element.action is action:
                for connection in canvas.list_connections(
                        self._p_object, group_id=element.group_id):
                    if connection.matches(
                            self._p_object.group_id, self._p_object.get_port_ids(),
                            element.group_id, element.port_id_list):
                        canvas_callback(
                            CallbackAct.PORTS_DISCONNECT,
                            connection.connection_id)
                break

    def add_element(self, group_id: int, port_id_list: list,
                    portgrp_id: int):
        if not port_id_list:
            return

        for element in self._elements:
            if (element.group_id == group_id
                    and element.port_id_list == port_id_list):
                # element already exists
                return

        # display actions in the group_id and port_id order
        i = 0
        following_action = None

        for element in self._elements:
            if (element.group_id > group_id
                    or (element.group_id == group_id
                        and element.port_id_list[0] > port_id_list[-1])):
                following_action = element.action
                break
            i += 1

        action_name = ""
        if isinstance(self._p_object, PortgrpObject) and portgrp_id:
            action_name = '‖ '
            action_name += get_portgroup_full_name(group_id, portgrp_id)
        else:
            action_name = get_full_port_name(group_id, port_id_list[0])

        icon = get_group_icon(group_id, self._p_object.port_mode.opposite(), is_dark_theme(self))

        action = QAction(action_name)
        action.setIcon(icon)
        action.triggered.connect(self.apply_disconnection)

        if following_action is None:
            self.addAction(action)
        else:
            self.insertAction(following_action, action)

        element = _DataDisconnElement()
        element.group_id = group_id
        element.portgrp_id = portgrp_id
        element.port_id_list = port_id_list
        element.action = action
        self._elements.insert(i, element)

        # remove "no connections" fake action
        if self._no_action is not None:
            self.removeAction(self._no_action)
            self._no_action = None

    def remove_element(self, group_id: int, port_id_list: int,
                       portgrp_id: int):
        for element in self._elements:
            if (element.group_id == group_id
                    and element.portgrp_id == portgrp_id
                    and element.port_id_list == port_id_list):
                self.removeAction(element.action)
                self._elements.remove(element)
                break
        else:
            return

        if not self._elements:
            self._no_action = self.addAction(self._no_action_title)
            self._no_action.setEnabled(False)


class ClipboardMenu(SubMenu):
    def __init__(self, p_object: Union[PortObject, PortgrpObject], parent):
        SubMenu.__init__(self, _translate('patchbay', 'Clipboard'),
                         p_object, parent)

        cut_action = self.addAction(
            _translate('patchbay', 'Cut connections'))
        cut_action.setIcon(QIcon.fromTheme('edit-cut'))
        cut_action.triggered.connect(self.cut_connections)

        copy_action = self.addAction(
            _translate('patchbay', 'Copy connections'))
        copy_action.setIcon(QIcon.fromTheme('edit-copy'))
        copy_action.triggered.connect(self.copy_connections)

        po = self._p_object

        for connection in canvas.list_connections(po):
            break
        else:
            cut_action.setEnabled(False)
            copy_action.setEnabled(False)

        for cb_element in canvas.clipboard:
            if (cb_element.port_type is po.port_type
                    and cb_element.port_mode is po.port_mode
                    and cb_element.port_id not in po.get_port_ids()):
                paste_action = self.addAction(
                    _translate('patchbay', 'Paste connections'))
                paste_action.setIcon(QIcon.fromTheme('edit-paste'))
                paste_action.triggered.connect(self.paste_connections)
                break

    def write_clipboard(self):
        canvas.clipboard.clear()

        po = self._p_object

        for self_port_id in po.get_port_ids():
            group_port_ids = list[tuple[int]]()

            if po.port_mode is PortMode.OUTPUT:
                for connection in canvas.list_connections(po):
                    if connection.port_out_id == self_port_id:
                        group_port_ids.append((connection.group_in_id,
                                               connection.port_in_id))
            elif po.port_mode is PortMode.INPUT:
                for connection in canvas.list_connections(po):
                    if connection.port_in_id == self_port_id:
                        group_port_ids.append((connection.group_out_id,
                                               connection.port_out_id))

            element = ClipboardElement()
            element.group_id = po.group_id
            element.port_id = self_port_id
            element.port_type = po.port_type
            element.port_mode = po.port_mode
            element.group_port_ids = group_port_ids

            canvas.clipboard.append(element)

    def cut_connections(self):
        canvas.clipboard_cut = True
        self.write_clipboard()

    def copy_connections(self):
        canvas.clipboard_cut = False
        self.write_clipboard()

    def paste_connections(self):
        po = self._p_object
        
        for i in range(len(po.get_port_ids())):
            for j in range(len(canvas.clipboard)):
                if i % len(canvas.clipboard) != j % len(po.get_port_ids()):
                    continue

                self_port_id = po.get_port_ids()[i]
                element = canvas.clipboard[j]

                if (element.port_type is po.port_type
                        and element.port_mode is po.port_mode):
                    for group_port_id in element.group_port_ids:
                        group_id, port_id = group_port_id

                        if canvas.clipboard_cut:
                            # remove the original connection if still exists
                            for connection in canvas.list_connections():
                                if connection.matches(
                                        element.group_id, [element.port_id],
                                        group_id, [port_id]):
                                    canvas_callback(
                                        CallbackAct.PORTS_DISCONNECT,
                                        connection.connection_id)
                                    break

                        connect_ports(po.group_id, self_port_id,
                                      group_id, port_id)
                    break

        # once paste, de-activate cut to prevent recut of connections
        # if they have been remade by user
        canvas.clipboard_cut = False


class ConnectableContextMenu(QMenu):
    def __init__(self, p_object: Union[PortObject, PortgrpObject]):
        QMenu.__init__(self)

        self._p_object = p_object
        po = self._p_object

        self.connection_list = list[ConnectionObject]()

        canvas.qobject.connection_added.connect(
            self.connection_added_to_canvas)
        canvas.qobject.connection_removed.connect(
            self.connection_removed_from_canvas)

        dark = '-dark' if is_dark_theme(self) else ''

        self.connect_menu = ConnectMenu(self._p_object, self)
        self.connect_menu.setIcon(
            QIcon(QPixmap(':scalable/breeze%s/lines-connector' % dark)))
        self.addMenu(self.connect_menu)

        self.disconnect_menu = DisconnectMenu(self._p_object, self)
        self.disconnect_menu.setIcon(
            QIcon(QPixmap(':scalable/breeze%s/lines-disconnector' % dark)))
        self.addMenu(self.disconnect_menu)

        disconnect_all_action = self.addAction(
            _translate('patchbay', "Disconnect All"))
        disconnect_all_action.setIcon(
            QIcon(QPixmap(':scalable/breeze%s/lines-disconnector' % dark)))
        disconnect_all_action.triggered.connect(self.disconnect_all)

        self.clipboard_menu = ClipboardMenu(self._p_object, self)
        self.clipboard_menu.setIcon(QIcon.fromTheme('edit-paste'))
        self.addMenu(self.clipboard_menu)

        self.addSeparator()

        for connection in canvas.list_connections(po):
            self.add_connection(connection)

    def disconnect_all(self):
        for connection in self.connection_list:
            canvas_callback(CallbackAct.PORTS_DISCONNECT,
                            connection.connection_id)

    def add_connection(self, connection: ConnectionObject):
        self.connection_list.append(connection)

        po = self._p_object

        for port in canvas.list_ports():
            if ((po.port_mode is PortMode.OUTPUT
                        and port.group_id == connection.group_in_id
                        and port.port_id == connection.port_in_id)
                    or (po.port_mode is PortMode.INPUT
                        and port.group_id == connection.group_out_id
                        and port.port_id == connection.port_out_id)):
                group_id = port.group_id
                port_id = port.port_id
                portgrp_id = port.portgrp_id
                port_id_list = [port_id]

                if isinstance(po, PortgrpObject) and portgrp_id:
                    port_id = -1
                    port_id_list = get_portgroup_port_list(
                        group_id, portgrp_id)

                con_state = get_portgroup_connection_state(
                    po.group_id, po.get_port_ids(),
                    group_id, port_id_list)

                for group_menu in self.connect_menu.group_menus:
                    if group_menu.group_id() == group_id:
                        group_menu.check_element(
                            port_id, portgrp_id, con_state)
                        break

                for group_menu in self.connect_menu.dangerous_submenu.group_menus:
                    if group_menu.group_id() == group_id:
                        group_menu.check_element(
                            port_id, portgrp_id, con_state)
                        break

                self.disconnect_menu.add_element(group_id, port_id_list,
                                                 portgrp_id)
                break

    def connection_added_to_canvas(self, connection_id: int):
        connection = canvas.get_connection(connection_id)
        if connection is not None:
            self.add_connection(connection)

    def connection_removed_from_canvas(self, connection_id: int):
        po = self._p_object
        
        for connection in self.connection_list:
            if connection.connection_id == connection_id:
                for port in canvas.list_ports():
                    if ((po.port_mode is PortMode.OUTPUT
                            and port.group_id == connection.group_in_id
                            and port.port_id == connection.port_in_id)
                        or (po.port_mode is PortMode.INPUT
                            and port.group_id == connection.group_out_id
                            and port.port_id == connection.port_out_id)):
                        group_id = port.group_id
                        port_id = port.port_id
                        portgrp_id = port.portgrp_id

                        if isinstance(po, PortgrpObject) and portgrp_id:
                            port_id = -1
                            port_id_list = get_portgroup_port_list(
                                group_id, portgrp_id)
                        else:
                            port_id_list = [port_id]

                        con_state = get_portgroup_connection_state(
                            po.group_id, po.get_port_ids(),
                            group_id, port_id_list)

                        for group_menu in self.connect_menu.group_menus:
                            if group_menu.group_id() == group_id:
                                group_menu.check_element(
                                    port_id, portgrp_id, con_state)
                                break

                        for group_menu in self.connect_menu.dangerous_submenu.group_menus:
                            if group_menu.group_id() == group_id:
                                group_menu.check_element(
                                    port_id, portgrp_id, con_state)
                                break

                        self.disconnect_menu.remove_element(
                            group_id, port_id_list, portgrp_id)
                        break

                self.connection_list.remove(connection)
                break
                



import json
from pathlib import Path
from typing import TYPE_CHECKING, Any
import os
import sys

import ray
import xdg
from gui_server_thread import GuiServerThread
from gui_tools import RS, get_code_root

from jack_renaming_tools import group_belongs_to_client
from patchbay.base_elements import (Group, GroupPos, PortgroupMem,
                                    PortMode, BoxLayoutMode, PortType, ToolDisplayed,
                                    PortTypesViewFlag)
from patchbay import (
    PatchbayManager,
    Callbacker,
    PatchbayToolsWidget,
    CanvasOptionsDialog,
    CanvasMenu,
    patchcanvas
)

if TYPE_CHECKING:
    from gui_session import Session
    from main_window import MainWindow


def convert_group_pos_from_ray_to_patchbay(
        ray_gpos: ray.GroupPosition) -> GroupPos:
    gpos = GroupPos()
    gpos.port_types_view = PortTypesViewFlag(ray_gpos.port_types_view)
    gpos.group_name = ray_gpos.group_name
    gpos.null_zone = ray_gpos.null_zone
    gpos.in_zone = ray_gpos.in_zone
    gpos.out_zone = ray_gpos.out_zone
    gpos.null_xy = ray_gpos.null_xy
    gpos.in_xy = ray_gpos.in_xy
    gpos.out_xy = ray_gpos.out_xy
    gpos.flags = ray_gpos.flags
    
    for port_mode in (PortMode.INPUT, PortMode.OUTPUT, PortMode.BOTH):
        layout_mode = ray_gpos.get_layout_mode(port_mode.value)
        gpos.set_layout_mode(port_mode, BoxLayoutMode(layout_mode))

    gpos.fully_set = ray_gpos.fully_set
    return gpos

def convert_group_pos_from_patchbay_to_ray(
        gpos: GroupPos) -> ray.GroupPosition:
    ray_gpos = ray.GroupPosition.new_from(
        int(gpos.port_types_view), gpos.group_name,
        gpos.null_zone, gpos.in_zone, gpos.out_zone,
        *gpos.null_xy, *gpos.in_xy, *gpos.out_xy,
        int(gpos.flags), 0)
    
    for port_mode, box_layout_mode in gpos.layout_modes.items():
        ray_gpos.set_layout_mode(port_mode.value, box_layout_mode.value)
    ray_gpos.fully_set = gpos.fully_set

    return ray_gpos

def convert_portgrp_mem_from_ray_to_patchbay(
        ray_pgmem: ray.PortGroupMemory) -> PortgroupMem:
    pgmem = PortgroupMem()
    pgmem.group_name = ray_pgmem.group_name
    try:
        pgmem.port_mode = PortMode(ray_pgmem.port_mode)
    except ValueError:
        pgmem.port_mode = PortMode.NULL

    try:
        pgmem.port_type = PortType(ray_pgmem.port_type)
    except ValueError:
        pgmem.port_type = PortType.NULL
    
    pgmem.above_metadatas = bool(ray_pgmem.above_metadatas) 
    # TODO, see if copy is required, probably not
    pgmem.port_names = ray_pgmem.port_names
    return pgmem

def convert_portgrp_mem_from_patchbay_to_ray(
        pgmem: PortgroupMem) -> ray.PortGroupMemory:
    return ray.PortGroupMemory.new_from(
        pgmem.group_name, pgmem.port_type.value, pgmem.port_mode.value,
        int(pgmem.above_metadatas), *pgmem.port_names)


class RayPatchbayCallbacker(Callbacker):
    def __init__(self, manager: 'RayPatchbayManager'):
        super().__init__(manager)
        self.mng = manager
        
    def _ports_connect(self, group_out_id: int, port_out_id: int,
                       group_in_id: int, port_in_id: int):
        port_out = self.mng.get_port_from_id(group_out_id, port_out_id)
        port_in = self.mng.get_port_from_id(group_in_id, port_in_id)

        if port_out is None or port_in is None:
            return

        self.mng.send_to_patchbay_daemon(
            '/ray/patchbay/connect',
            port_out.full_name, port_in.full_name)

    def _ports_disconnect(self, connection_id: int):
        for connection in self.mng.connections:
            if connection.connection_id == connection_id:
                self.mng.send_to_patchbay_daemon(
                    '/ray/patchbay/disconnect',
                    connection.port_out.full_name,
                    connection.port_in.full_name)
                break

    def _client_show_gui(self, group_id: int, visible: int):
        group = self.mng.get_group_from_id(group_id)
        if group is None:
            return

        for client in self.mng.session.client_list:
            if group_belongs_to_client(group.name, client.jack_client_name):
                show = 'show' if visible else 'hide'
                self.mng.send_to_daemon(
                    f'/ray/client/{show}_optional_gui',
                    client.client_id)
                break
            
    def _group_selected(self, group_id: int, splitted_mode: PortMode):
        # select client widget matching with the selected box
        group = self.mng.get_group_from_id(group_id)
        if group is None:
            return
        
        for client in self.mng.session.client_list:
            if group_belongs_to_client(group.name, client.jack_client_name):
                item = client.widget.list_widget_item
                item.setSelected(True)
                client.widget.list_widget.scrollToItem(item)
                break                


class RayPatchbayManager(PatchbayManager):    
    def __init__(self, session: 'Session'):
        super().__init__(RS.settings)
        self.session = session
        self.set_tools_widget(PatchbayToolsWidget())
        
        self._last_selected_client_name = ''
        self._last_selected_box_n = 1

    @staticmethod
    def send_to_patchbay_daemon(*args):
        server = GuiServerThread.instance()
        if not server:
            return

        if server.patchbay_addr is None:
            return

        server.send(server.patchbay_addr, *args)

    @staticmethod
    def send_to_daemon(*args):
        server = GuiServerThread.instance()
        if not server:
            return
        server.to_daemon(*args)
    
    def _get_json_contents_from_path(self, file_path: str) -> dict:
        if not os.path.exists(file_path):
            return {}

        if not os.access(file_path, os.R_OK):
            return {}

        try:
            file = open(file_path, 'r')
        except IOError:
            return {}

        try:
            new_dict = json.load(file)
            assert isinstance(new_dict, dict)
        except ImportError:
            return {}

        file.close()
        return new_dict
    
    def _setup_canvas(self):
        SUBMODULE = 'HoustonPatchbay'
        source_theme_path = get_code_root() / SUBMODULE / 'themes'
        manual_path = get_code_root() / SUBMODULE / 'manual'
        theme_paths = list[Path]()
        
        app_title = ray.APP_TITLE.lower()
        
        theme_paths.append(xdg.xdg_data_home() / app_title / SUBMODULE / 'themes')

        if source_theme_path.exists():
            theme_paths.append(source_theme_path)

        for p in xdg.xdg_data_dirs():
            path = p / app_title / SUBMODULE / 'themes'
            if path not in theme_paths:
                theme_paths.append(path)

        if TYPE_CHECKING:
            assert isinstance(self.main_win, MainWindow)

        self.app_init(self.main_win.ui.graphicsView,
                      theme_paths,
                      callbacker=RayPatchbayCallbacker(self),
                      manual_path=manual_path,
                      default_theme_name='Yellow Boards')
    
    #### reimplemented functions ###
    
    def refresh(self):
        super().refresh()
        self.send_to_patchbay_daemon('/ray/patchbay/refresh')
    
    def disannounce(self):
        self.send_to_patchbay_daemon('/ray/patchbay/gui_disannounce')
        super().disannounce()
    
    def save_group_position(self, gpos: GroupPos):
        super().save_group_position(gpos)
        ray_gpos = convert_group_pos_from_patchbay_to_ray(gpos)
        self.send_to_daemon(
            '/ray/server/patchbay/save_group_position', *ray_gpos.spread())

    def save_portgroup_memory(self, portgrp_mem: PortgroupMem):
        super().save_portgroup_memory(portgrp_mem)
        ray_pgmem = convert_portgrp_mem_from_patchbay_to_ray(portgrp_mem)
        self.send_to_daemon(
            '/ray/server/patchbay/save_portgroup',
            *ray_pgmem.spread())

    def change_buffersize(self, buffer_size: int):
        super().change_buffersize(buffer_size)
        self.send_to_patchbay_daemon('/ray/patchbay/set_buffer_size',
                                     buffer_size)

    def filter_groups(self, text: str, n_select=0) -> int:
        ''' semi hides groups not matching with text
            and returns number of matching boxes '''
        opac_grp_ids = set()
        opac_conn_ids = set()
        
        if text.startswith(('cl:', 'client:')):
            client_ids = text.rpartition(':')[2].split(' ')
            jack_client_names = list[str]()
            
            for client in self.session.client_list:
                if (client.status != ray.ClientStatus.STOPPED
                        and client.client_id in client_ids):
                    jack_client_names.append(client.jack_client_name)
                    if not client.jack_client_name.endswith('.' + client.client_id):
                        jack_client_names.append(client.jack_client_name + '.0')
            
            for group in self.groups:
                opac = False
                for jack_client_name in jack_client_names:
                    if (group.name == jack_client_name
                            or group.name.startswith(jack_client_name + '/')
                            or (group.name.startswith(jack_client_name + ' (')
                                    and ')' in group.name)):
                        break
                else:
                    opac = True
                    opac_grp_ids.add(group.group_id)
                
                group.semi_hide(opac)

        else:
            for group in self.groups:
                opac = bool(text.lower() not in group.name.lower()
                            and text.lower() not in group.display_name.lower())
                if opac:
                    opac_grp_ids.add(group.group_id)

                group.semi_hide(opac)
        
        for conn in self.connections:
            opac_conn = bool(
                conn.port_out.group_id in opac_grp_ids
                and conn.port_in.group_id in opac_grp_ids)
            
            conn.semi_hide(opac_conn)
            if opac_conn:
                opac_conn_ids.add(conn.connection_id)
            
        for group in self.groups:
            if group.group_id in opac_grp_ids:
                group.set_in_front()
        
        for conn in self.connections:
            if conn.connection_id in opac_conn_ids:
                conn.set_in_front()
        
        for conn in self.connections:
            if conn.connection_id not in opac_conn_ids:
                conn.set_in_front()
        
        n_boxes = 0
        
        for group in self.groups:
            if group.group_id not in opac_grp_ids:
                group.set_in_front()
                n_grp_boxes = group.get_number_of_boxes()

                if n_select > n_boxes and n_select <= n_boxes + n_grp_boxes:
                    group.select_filtered_box(n_select - n_boxes)
                n_boxes += n_grp_boxes

        return n_boxes

    def get_corrected_a2j_group_name(self, group_name: str) -> str:
        # fix a2j wrongly substitute '.' with space
        for client in self.session.client_list:
            if (client.status != ray.ClientStatus.STOPPED
                    and '.' in client.jack_client_name
                    and (client.jack_client_name.replace('.', ' ', 1)
                            == group_name)):
                return client.jack_client_name
        
        return group_name

    def set_group_as_nsm_client(self, group: Group):
        for client in self.session.client_list:
            if group_belongs_to_client(group.name, client.jack_client_name):
                group.set_client_icon(client.icon)
                
                # in case of long jack naming (ClientName.ClientId)
                # do not display ClientName if we have the icon
                if (client.icon
                        and client.jack_client_name.endswith('.' + client.client_id)
                        and group.name.startswith(client.jack_client_name)):
                    group.display_name = group.display_name.partition('.')[2]
                
                if client.has_gui:
                    group.set_optional_gui_state(client.gui_state)
                break

    def transport_play_pause(self, play: bool):
        self.send_to_patchbay_daemon('/ray/patchbay/transport_play', int(play))
    
    def transport_stop(self):
        self.send_to_patchbay_daemon('/ray/patchbay/transport_stop')
    
    def transport_relocate(self, frame: int):
        self.send_to_patchbay_daemon('/ray/patchbay/transport_relocate', frame)

    def change_tools_displayed(self, tools_displayed: ToolDisplayed):
        ex_tool_displayed = self._tools_displayed
        super().change_tools_displayed(tools_displayed)
        ex_transport_disp = ex_tool_displayed & (ToolDisplayed.TRANSPORT_CLOCK
                                                 | ToolDisplayed.TRANSPORT_PLAY_STOP)
        transport_disp = tools_displayed & (ToolDisplayed.TRANSPORT_CLOCK
                                           | ToolDisplayed.TRANSPORT_PLAY_STOP)

        if transport_disp != ex_transport_disp:
            if transport_disp & ToolDisplayed.TRANSPORT_CLOCK:
                int_transp = 2
            elif transport_disp & ToolDisplayed.TRANSPORT_PLAY_STOP:
                int_transp = 1
            else:
                int_transp = 0

            self.send_to_patchbay_daemon('/ray/patchbay/activate_transport',
                                          int_transp)
             
        if (ex_tool_displayed & ToolDisplayed.DSP_LOAD
                != tools_displayed & ToolDisplayed.DSP_LOAD):
            self.send_to_patchbay_daemon(
                '/ray/patchbay/activate_dsp_load',
                int(bool(tools_displayed & ToolDisplayed.DSP_LOAD)))

        #### added functions ####
    
    def select_client_box(self, jack_client_name: str, previous=False):
        if not jack_client_name:
            self._last_selected_client_name = ''
            self._last_selected_box_n = 0
            patchcanvas.canvas.scene.clearSelection()
            return
        
        box_n = 0
        if jack_client_name == self._last_selected_client_name:            
            n_max = 0
            for group in self.groups:
                if group_belongs_to_client(group.name, jack_client_name):
                    n_max += group.get_number_of_boxes()
            
            if previous:
                if self._last_selected_box_n == 0:
                    self._last_selected_box_n = n_max -1
                else:
                    self._last_selected_box_n -= 1
            else:
                self._last_selected_box_n += 1

            if n_max:
                box_n = self._last_selected_box_n % n_max
            
        n = 0
        box_found = False
        for group in self.groups:
            if group_belongs_to_client(group.name, jack_client_name):
                if group.get_number_of_boxes() + n <= box_n:
                    n += group.get_number_of_boxes()
                    continue
                
                group.select_filtered_box(1 + box_n -n)
                box_found = True
                break
        
        if box_found:
            self._last_selected_client_name = jack_client_name
            self._last_selected_box_n = box_n
        else:
            patchcanvas.canvas.scene.clearSelection()
            self._last_selected_client_name = ''
            self._last_selected_box_n = 0

    def update_group_position(self, *args):
        # arguments are these ones delivered from ray.GroupPosition.spread()
        # Not define them allows easier code modifications.
        gpos = convert_group_pos_from_ray_to_patchbay(
            ray.GroupPosition.new_from(*args))
        
        for gposition in self.group_positions:
            if (gposition.group_name == gpos.group_name
                    and gposition.port_types_view == gpos.port_types_view):
                gposition.eat(gpos)
                break
        else:
            self.group_positions.append(gpos)
        
        if gpos.port_types_view == self.port_types_view:
            group = self.get_group_from_name(gpos.group_name)
            if group is not None:
                group.set_group_position(gpos)

    def update_portgroup(self, *args):
        ray_pgmem = ray.PortGroupMemory.new_from(*args)
        pg_mem = convert_portgrp_mem_from_ray_to_patchbay(ray_pgmem)
        self.add_portgroup_memory(pg_mem)

        group = self.get_group_from_name(pg_mem.group_name)
        if group is not None:
            group.portgroup_memory_added(pg_mem)
    
    def optional_gui_state_changed(self, client_id: str, visible: bool):
        for client in self.session.client_list:
            if client.client_id == client_id:
                for group in self.groups:
                    if group_belongs_to_client(group.name, client.jack_client_name):
                        group.set_optional_gui_state(visible)
                break
    
    def receive_big_packets(self, state: int):
        self.optimize_operation(not bool(state))
        if state:
            self.redraw_all_groups()
            
    def finish_init(self):
        self.set_main_win(self.session.main_win)
        self._setup_canvas()
        self.set_canvas_menu(CanvasMenu(self))
        self.set_options_dialog(
            CanvasOptionsDialog(self.main_win, self, RS.settings))
        
    def fast_temp_file_memory(self, temp_path):
        '''receives a .json file path from daemon with groups positions
           and portgroups remembered from user.'''
        canvas_data = self._get_json_contents_from_path(temp_path)
        if not canvas_data:
            sys.stderr.write(
                "RaySession::Failed to load tmp file %s to get canvas positions\n"
                % temp_path)
            return

        for key in canvas_data.keys():
            if key == 'group_positions':
                for gpos_dict in canvas_data[key]:
                    gpos = ray.GroupPosition()
                    gpos.write_from_dict(gpos_dict)
                    self.update_group_position(*gpos.spread())

            elif key == 'portgroups':
                for pg_dict in canvas_data[key]:
                    portgroup_mem = ray.PortGroupMemory()
                    portgroup_mem.write_from_dict(pg_dict)
                    self.update_portgroup(*portgroup_mem.spread())

        try:
            os.remove(temp_path)
        except:
            pass

    def fast_temp_file_running(self, temp_path: str):
        '''receives a .json file path from patchbay daemon with all ports, connections
           and jack metadatas'''
            
        patchbay_data = self._get_json_contents_from_path(temp_path)
        if not patchbay_data:
            sys.stderr.write(
                "RaySession::Failed to load tmp file %s to get JACK ports\n"
                % temp_path)
            return

        self.clear_all()

        # optimize_operation allow to not redraw group at each port added.
        # however, if there is no group position
        # (i.e. if there is no config at all), it is prefferable to
        # know where finish the group boxes before to add another one.
        
        # very fast operation means that nothing is done in the patchcanvas
        # everything stays here in this file.
        if self.group_positions:
            self.optimize_operation(True)
            self._set_very_fast_operation(True)

        for key in patchbay_data.keys():
            if key == 'ports':
                p: dict[str, Any]
                for p in patchbay_data[key]:
                    self.add_port(p.get('name'), p.get('type'),
                                  p.get('flags'), p.get('uuid'))

            elif key == 'connections':
                c: dict[str, Any]
                for c in patchbay_data[key]:
                    self.add_connection(c.get('port_out_name'),
                                        c.get('port_in_name'))

        for key in patchbay_data.keys():
            if key == 'clients':
                cnu: dict[str, Any]
                for cnu in patchbay_data[key]:
                    self.set_group_uuid_from_name(cnu.get('name'), cnu.get('uuid'))
                break

        for key in patchbay_data.keys():
            if key == 'metadatas':
                m: dict[str, Any]
                for m in patchbay_data[key]:
                    self.metadata_update(
                        m.get('uuid'), m.get('key'), m.get('value'))
                break

        for group in self.groups:
            group.sort_ports_in_canvas()

        self._set_very_fast_operation(False)
                
        for group in self.groups:
            group.add_all_ports_to_canvas()
        
        for conn in self.connections:
            conn.add_to_canvas()

        self.optimize_operation(False)
        self.redraw_all_groups()

        try:
            os.remove(temp_path)
        except:
            # if this tmp file can not be removed
            # this is really not strong.
            pass

    def patchbay_announce(self, jack_running: int, samplerate: int,
                          buffer_size: int):
        if self._tools_widget is None:
            return
        
        self._tools_widget.set_samplerate(samplerate)
        self._tools_widget.set_buffer_size(buffer_size)
        self._tools_widget.set_jack_running(jack_running)

        if self.main_win is not None:
            if TYPE_CHECKING and not isinstance(self.main_win, MainWindow):
                return
            self.main_win.add_patchbay_tools(
                self._tools_widget, self.canvas_menu)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys

from ctypes import pointer
import jacklib
from jacklib import JackTransportState
from jacklib.helpers import get_jack_status_error_string


status = jacklib.jack_status_t()
client = jacklib.client_open(
    "transport-query", jacklib.JackOptions.NO_START_SERVER, status)
err = get_jack_status_error_string(status)

if status.value:
    if status.value & jacklib.JackStatus.NAME_NOT_UNIQUE:
        print("Non-fatal JACK status: %s" % err, file=sys.stderr)
    elif status.value & jacklib.JackStatus.SERVER_STARTED:
        # Should not happen, since we use the JackOptions.NO_START_SERVER option
        print("Unexpected JACK status: %s" % err, file=sys.stderr)
    else:
        sys.exit("Error connecting to JACK server: %s" % err)

position = jacklib.jack_position_t()
transport_state = jacklib.transport_query(client, pointer(position))

for name, type_ in sorted(position._fields_):
    value = getattr(position, name)
    if name == 'padding':
        value = list(value)
    print("{}: {}".format(name, value), file=sys.stderr)

if transport_state == JackTransportState.STOPPED:
    print("JACK transport stopped, starting it now.")
    jacklib.transport_start(client)
elif transport_state == JackTransportState.ROLLING:
    print("JACK transport rolling, stopping it now.")
    jacklib.transport_stop(client)
elif transport_state == JackTransportState.STARTING:
    print("JACK transport starting, nothing to do.")
else:
    print("Unknown JACK transport state.")

jacklib.client_close(client)




from PyQt5.QtCore import Qt, QPoint
from PyQt5.QtGui import QMouseEvent, QWheelEvent
from PyQt5.QtWidgets import QApplication, QGraphicsView

# taken from carla (falktx)
class PatchGraphicsView(QGraphicsView):
    def __init__(self, parent):
        super().__init__(parent)
        self.setTransformationAnchor(QGraphicsView.AnchorUnderMouse)

        self._panning = False

        try:
            self._middle_button = Qt.MiddleButton
        except:
            self._middle_button = Qt.MidButton

    def mousePressEvent(self, event):
        if (event.button() == self._middle_button
                and not QApplication.keyboardModifiers() & Qt.ControlModifier):
            self._panning = True
            self.setDragMode(QGraphicsView.ScrollHandDrag)
            event = QMouseEvent(event.type(), event.pos(), Qt.LeftButton,
                                Qt.LeftButton, event.modifiers())

        QGraphicsView.mousePressEvent(self, event)

    def mouseReleaseEvent(self, event):
        QGraphicsView.mouseReleaseEvent(self, event)

        if not self._panning:
            return

        self._panning = False
        self.setDragMode(QGraphicsView.NoDrag)
        
    def wheelEvent(self, ev: QWheelEvent) -> None:
        if ev.modifiers() == Qt.ShiftModifier:
            # lie to Qt saying to QGraphicsView and QGraphicsScene
            # that keyboard modifier key is ALT
            x, y = ev.angleDelta().x(), ev.angleDelta().y()
            new_delta = QPoint(y, x)
            new_event = QWheelEvent(
                ev.posF(), ev.globalPosF(), ev.pixelDelta(), new_delta,
                ev.buttons(), Qt.AltModifier, ev.phase(), ev.inverted(),
                ev.source())
            QGraphicsView.wheelEvent(self, new_event)
            return

        QGraphicsView.wheelEvent(self, ev)


# -*- coding: utf-8 -*-

from enum import IntEnum
import os
try:
    from liblo import Server, make_method, Address
except ImportError:
    from pyliblo3 import Server, make_method, Address
from typing import Callable, Optional


class Err(IntEnum):
    OK =  0
    GENERAL_ERROR = -1
    INCOMPATIBLE_API = -2
    BLACKLISTED = -3
    LAUNCH_FAILED = -4
    NO_SUCH_FILE = -5
    NO_SESSION_OPEN = -6
    UNSAVED_CHANGES = -7
    NOT_NOW = -8
    BAD_PROJECT = -9
    CREATE_FAILED = -10
    SESSION_LOCKED = -11
    OPERATION_PENDING = -12


class NsmCallback(IntEnum):
    OPEN = 1
    SAVE = 2
    SESSION_IS_LOADED = 3
    SHOW_OPTIONAL_GUI = 4
    HIDE_OPTIONAL_GUI = 5
    MONITOR_CLIENT_STATE = 6
    MONITOR_CLIENT_EVENT = 7
    MONITOR_CLIENT_UPDATED = 8


class NsmServer(Server):
    def __init__(self, daemon_address: Address):
        Server.__init__(self)
        self._daemon_address = daemon_address
        self._server_capabilities = ""
        
        self._callbacks = dict[NsmCallback, Callable]()

    @make_method('/reply', None)
    def _reply(self, path, args):
        if args:
            reply_path = args[0]
        else:
            return

        if reply_path == '/nsm/server/announce':
            self._server_capabilities = args[3]

    @make_method('/nsm/client/open', 'sss')
    def _nsm_client_open(self, path, args):
        ret = self._exec_callback(NsmCallback.OPEN, *args)
        if ret is None:
            return
        
        err, err_text = ret
        if err is Err.OK:
            self._send_to_daemon('/reply', path, 'Ready')
        else:
            self._send_to_daemon('/error', path, err, err_text)

    @make_method('/nsm/client/save', '')
    def _nsm_client_save(self, path, args):
        ret = self._exec_callback(NsmCallback.SAVE)
        if ret is None:
            return
        
        err, err_text = ret
        if err is Err.OK:
            self._send_to_daemon('/reply', path, 'Saved')
        else:
            self._send_to_daemon('/error', path, err_text)

    @make_method('/nsm/client/session_is_loaded', '')
    def _nsm_client_session_is_loaded(self, path, args):
        self._exec_callback(NsmCallback.SESSION_IS_LOADED)

    @make_method('/nsm/client/show_optional_gui', '')
    def _nsm_client_show_optional_gui(self, path, args):
        self._exec_callback(NsmCallback.SHOW_OPTIONAL_GUI)

    @make_method('/nsm/client/hide_optional_gui', '')
    def _nsm_client_hide_optional_gui(self, path, args):
        self._exec_callback(NsmCallback.HIDE_OPTIONAL_GUI)
    
    @make_method('/nsm/client/monitor/client_state', 'ssi')
    def _nsm_client_monitor_client_state(self, path, args):
        self._exec_callback(NsmCallback.MONITOR_CLIENT_STATE, *args)
    
    @make_method('/nsm/client/monitor/client_event', 'ss')
    def _nsm_client_monitor_client_event(self, path, args):
        self._exec_callback(NsmCallback.MONITOR_CLIENT_EVENT, *args)

    @make_method('/nsm/client/monitor/client_updated', 'ssi')
    def _nsm_client_monitor_client_properties(self, path, args):
        self._exec_callback(NsmCallback.MONITOR_CLIENT_UPDATED, *args)
    
    def set_callback(self, on_event: NsmCallback, func: Callable):
        self._callbacks[on_event] = func

    def _exec_callback(self, event: NsmCallback, *args) -> Optional[tuple[Err, str]]:
        if event in self._callbacks.keys():
            return self._callbacks[event](*args)

    def get_server_capabilities(self):
        return self._server_capabilities

    def _send_to_daemon(self, *args):
        self.send(self._daemon_address, *args)

    def announce(self, client_name: str, capabilities: str, executable_path: str):
        MAJOR, MINOR = 1, 0
        pid = os.getpid()

        self._send_to_daemon(
            '/nsm/server/announce',
            client_name,
            capabilities,
            executable_path,
            MAJOR,
            MINOR,
            pid)

    def send_dirty_state(self, dirty: bool):
        if dirty:
            self._send_to_daemon('/nsm/client/is_dirty')
        else:
            self._send_to_daemon('/nsm/client/is_clean')

    def send_gui_state(self, state: bool):
        if state:
            self._send_to_daemon('/nsm/client/gui_is_shown')
        else:
            self._send_to_daemon('/nsm/client/gui_is_hidden')
            
    def send_monitor_reset(self):
        if ':monitor:' in self._server_capabilities:
            self._send_to_daemon('/nsm/server/monitor_reset')





from .patchbay_manager import PatchbayManager, patchcanvas
from .base_elements import Group, Port, Portgroup
from .calbacker import Callbacker
from .tools_widgets import CanvasPortInfoDialog, CanvasMenu, PatchbayToolsWidget
from .options_dialog import CanvasOptionsDialog
from .filter_frame import FilterFrame


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys

import jacklib
from jacklib.helpers import get_jack_status_error_string


if len(sys.argv) == 3:
    portname = sys.argv[1]
    pretty_name = sys.argv[2]
else:
    sys.exit("Usage: %s <port name> <pretty-name>" % sys.argv[0])

status = jacklib.jack_status_t()
client = jacklib.client_open("set-port-pretty-name", jacklib.JackOptions.NO_START_SERVER, status)
err = get_jack_status_error_string(status)

if status.value:
    if status.value & jacklib.JackStatus.NAME_NOT_UNIQUE:
        print("Non-fatal JACK status: %s" % err, file=sys.stderr)
    elif status.value & jacklib.JackStatus.SERVER_STARTED:
        # Should not happen, since we use the JackOptions.NO_START_SERVER option
        print("Unexpected JACK status: %s" % err, file=sys.stderr)
    else:
        sys.exit("Error connecting to JACK server: %s" % err)

res = jacklib.set_port_pretty_name(client, portname, pretty_name)
if res != -1:
    print("Pretty name for port '%s' is now '%s'." % (portname, pretty_name))


jacklib.client_close(client)




import json
import logging
import operator
from pathlib import Path
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Callable, Optional, Union
from unittest.mock import patch

from PyQt5.QtGui import QCursor, QGuiApplication
from PyQt5.QtWidgets import QMainWindow, QMessageBox, QWidget
from PyQt5.QtCore import QTimer, QSettings, QThread, QTranslator, QLocale

from .patchcanvas import patchcanvas, PortType, PortSubType, PortMode
from .patchcanvas.utils import get_new_group_positions
from .patchcanvas.scene_view import PatchGraphicsView
from .patchcanvas.init_values import (
    CallbackAct, CanvasFeaturesObject, CanvasOptionsObject)

from .patchbay_signals import SignalsObject
from .tools_widgets import PatchbayToolsWidget
from .canvas_menu import CanvasMenu
from .options_dialog import CanvasOptionsDialog
from .filter_frame import FilterFrame
from .base_elements import (
    Connection, GroupPos, Port, PortTypesViewFlag, Portgroup, Group,
    JackPortFlag, PortgroupMem, ToolDisplayed, TransportPosition)
from .calbacker import Callbacker


# Meta data (taken from pyjacklib)
_JACK_METADATA_PREFIX = "http://jackaudio.org/metadata/"
JACK_METADATA_CONNECTED = _JACK_METADATA_PREFIX + "connected"
JACK_METADATA_EVENT_TYPES = _JACK_METADATA_PREFIX + "event-types"
JACK_METADATA_HARDWARE = _JACK_METADATA_PREFIX + "hardware"
JACK_METADATA_ICON_LARGE = _JACK_METADATA_PREFIX + "icon-large"
JACK_METADATA_ICON_NAME = _JACK_METADATA_PREFIX + "icon-name"
JACK_METADATA_ICON_SMALL = _JACK_METADATA_PREFIX + "icon-small"
JACK_METADATA_ORDER = _JACK_METADATA_PREFIX + "order"
JACK_METADATA_PORT_GROUP = _JACK_METADATA_PREFIX + "port-group"
JACK_METADATA_PRETTY_NAME = _JACK_METADATA_PREFIX + "pretty-name"
JACK_METADATA_SIGNAL_TYPE = _JACK_METADATA_PREFIX + "signal-type"

_translate = QGuiApplication.translate
_logger = logging.getLogger(__name__)

def enum_to_flag(enum_int: int) -> int:
    if enum_int <= 0:
        return 0
    return 2 ** (enum_int - 1)

@dataclass
class DelayedOrder:
    func: Callable
    args: tuple
    kwargs: dict
    draw_group: bool
    sort_group: bool
    clear_conns: bool

    
def later_by_batch(draw_group=False, sort_group=False, clear_conns=False):
    def decorator(func: Callable):
        def wrapper(*args, **kwargs):
            mng = args[0]
            assert isinstance(mng, PatchbayManager)
            if mng.very_fast_operation:
                return func(*args, **kwargs)

            mng.delayed_orders.append(
                DelayedOrder(func, args, kwargs,
                             draw_group or sort_group,
                             sort_group,
                             clear_conns))

            if QThread.currentThread() is QGuiApplication.instance().thread():
                mng._delayed_orders_timer.start()
            else:
                mng.sg.out_thread_order.emit()
            return
        return wrapper
    return decorator


def in_main_thread():
    def decorator(func: Callable):
        def wrapper(*args, **kwargs):
            mng = args[0]
            assert isinstance(mng, PatchbayManager)
            
            if QThread.currentThread() is QGuiApplication.instance().thread():
                return func(*args, **kwargs)
            
            mng.sg.to_main_thread.emit(func, args, kwargs)
        return wrapper
    return decorator


class PatchbayManager:
    use_graceful_names = True
    port_types_view = PortTypesViewFlag.ALL
    
    # when True, items can be added to patchcanvas but they won't be drawn
    optimized_operation = False
    
    # when True, items are not added/removed from patchcanvas
    # useful to win time at startup or refresh
    very_fast_operation = False

    groups = list[Group]()
    connections = list[Connection]()
    _groups_by_name = dict[str, Group]()
    _groups_by_id = dict[int, Group]()
    _ports_by_name = dict[str, Port]()

    group_positions = list[GroupPos]()
    portgroups_memory = list[PortgroupMem]()
    delayed_orders = list[DelayedOrder]()

    def __init__(self, settings: QSettings):
        self.callbacker = Callbacker(self)
        self._settings = settings

        self.main_win: QMainWindow = None
        self._tools_widget: PatchbayToolsWidget = None
        self.options_dialog: CanvasOptionsDialog = None
        self.filter_frame: FilterFrame = None
        
        self._manual_path: Path = None

        self.server_is_started = True

        self.sg = SignalsObject()

        self._next_group_id = 0
        self._next_port_id = 0
        self._next_connection_id = 0
        self._next_portgroup_id = 1

        self.set_graceful_names(settings.value(
            'Canvas/use_graceful_names', True, type=bool))
        self.group_a2j_hw = settings.value(
            'Canvas/group_a2j_ports', True, type=bool)
        self.alsa_midi_enabled = settings.value(
            'Canvas/alsa_midi_enabled', False, type=bool)

        # all patchbay events are delayed
        # to reduce the patchbay comsumption.
        # Redraws in canvas are made once 50ms have passed without any event.
        # This prevent one group redraw per port added/removed
        # when a lot of ports are added/removed/renamed simultaneously.
        self._delayed_orders_timer = QTimer()
        self._delayed_orders_timer.setInterval(50)
        self._delayed_orders_timer.setSingleShot(True)
        self._delayed_orders_timer.timeout.connect(
            self._delayed_orders_timeout)

        self.sg.out_thread_order.connect(self._delayed_orders_timer.start)
        self.sg.to_main_thread.connect(self._execute_in_main_thread)
        
        self.sg.theme_changed.connect(self.change_theme)
        
        self.sg.graceful_names_changed.connect(self.set_graceful_names)
        self.sg.a2j_grouped_changed.connect(self.set_a2j_grouped)
        self.sg.alsa_midi_enabled_changed.connect(self.set_alsa_midi_enabled)
        self.sg.group_shadows_changed.connect(self.set_group_shadows)
        self.sg.auto_select_items_changed.connect(self.set_auto_select_items)
        self.sg.elastic_changed.connect(self.set_elastic_canvas)
        self.sg.borders_nav_changed.connect(self.set_borders_navigation)
        self.sg.prevent_overlap_changed.connect(self.set_prevent_overlap)
        self.sg.max_port_width_changed.connect(patchcanvas.set_max_port_width)
        self.sg.default_zoom_changed.connect(patchcanvas.set_default_zoom)
        
        self._tools_displayed = ToolDisplayed.ALL

    def __canvas_callback__(self, action: CallbackAct, *args):
        self.sg.callback_sig.emit(action, args)

    def app_init(self,
                 view: PatchGraphicsView,
                 theme_paths: list[Path],
                 manual_path: Path = None,
                 callbacker: Callbacker = None,
                 options: CanvasOptionsObject = None,
                 features: CanvasFeaturesObject = None,
                 default_theme_name='Black Gold'):
        if callbacker is not None:
            if not isinstance(callbacker, Callbacker):
                exception = TypeError("callbacker must be a Callbacker instance !")
                raise exception
            
            self.callbacker = callbacker
            self.sg.callback_sig.connect(self.callbacker.receive)
        
        self._manual_path = manual_path
        
        if options is None:
            options = patchcanvas.CanvasOptionsObject()
            if isinstance(self._settings, QSettings):
                options.theme_name = self._settings.value(
                    'Canvas/theme', default_theme_name, type=str)
                options.show_shadows = self._settings.value(
                    'Canvas/box_shadows', False, type=bool)
                options.auto_select_items = self._settings.value(
                    'Canvas/auto_select_items', False, type=bool)
                options.inline_displays = False
                options.elastic = self._settings.value(
                    'Canvas/elastic', True, type=bool)
                options.borders_navigation = self._settings.value(
                    'Canvas/borders_navigation', True, type=bool)
                options.prevent_overlap = self._settings.value(
                    'Canvas/prevent_overlap', True, type=bool)
                options.max_port_width = self._settings.value(
                    'Canvas/max_port_width', 160, type=int)
                options.semi_hide_opacity = self._settings.value(
                    'Canvas/semi_hide_opacity', 0.17, type=float)
                options.default_zoom = self._settings.value(
                    'Canvas/default_zoom', 100, type=int)
                options.box_grouped_auto_layout_ratio = self._settings.value(
                    'Canvas/grouped_box_auto_layout_ratio', 1.0, type=float)
        
        if features is None:
            features = CanvasFeaturesObject()

        patchcanvas.set_options(options)
        patchcanvas.set_features(features)
        patchcanvas.init(view, self.__canvas_callback__, theme_paths, default_theme_name)
        patchcanvas.canvas.scene.scale_changed.connect(self._scene_scale_changed)
        
        # just to have the zoom slider updated with the default zoom
        patchcanvas.canvas.scene.zoom_reset()

        # get port_types_view from config file only if the port type view
        # filter widget is visible.         
        try:
            port_types_view = PortTypesViewFlag(
                self._settings.value(
                    'Canvas/default_port_types_view',
                    PortTypesViewFlag.ALL.value,
                    type=int))
        except:
            _logger.warning(
                "unable to find correct default port types view "
                "in the config file.")
            port_types_view = PortTypesViewFlag.ALL
                
        self.change_port_types_view(port_types_view)
        
    def _scene_scale_changed(self, value: float):
        self.sg.scene_scale_changed.emit(value)

    def _execute_in_main_thread(self, func: Callable, args: tuple, kwargs: dict):
        func(*args, **kwargs)

    # --- widgets related methods --- #

    def set_main_win(self, main_win: QWidget):
        self.main_win = main_win

    def set_tools_widget(self, tools_widget: PatchbayToolsWidget):
        self._tools_widget = tools_widget
        self._tools_widget.buffer_size_change_order.connect(
            self.change_buffersize)
        self._tools_widget.set_patchbay_manager(self)

    def set_canvas_menu(self, canvas_menu: CanvasMenu):
        self.canvas_menu = canvas_menu

    def set_filter_frame(self, filter_frame: FilterFrame):
        self.filter_frame = filter_frame
        self.filter_frame.set_settings(self._settings)

    def set_options_dialog(self, options_dialog: CanvasOptionsDialog):
        self.options_dialog = options_dialog

    def show_options_dialog(self):
        if self.options_dialog is None:
            return
        
        self.options_dialog.move(QCursor.pos())
        self.options_dialog.show()

    # --- manager methods --- #

    @staticmethod
    def save_patchcanvas_cache():
        patchcanvas.save_cache()

    def set_use_graceful_names(self, yesno: bool):
        self.use_graceful_names = yesno

    def optimize_operation(self, yesno: bool):
        self.optimized_operation = yesno
        if patchcanvas.canvas is not None:
            patchcanvas.set_loading_items(yesno)

    def _set_very_fast_operation(self, yesno: bool):
        self.very_fast_operation = yesno

    def _add_group(self, group: Group):
        self.groups.append(group)
        self._groups_by_id[group.group_id] = group
        self._groups_by_name[group.name] = group

    def _remove_group(self, group: Group):
        if group in self.groups:
            self.groups.remove(group)
            self._groups_by_id.pop(group.group_id)
            self._groups_by_name.pop(group.name)

    def _clear_groups(self):
        self.groups.clear()
        self._groups_by_id.clear()
        self._groups_by_name.clear()

    def new_portgroup(self, group_id: int, port_mode: int,
                      ports: tuple[Port]) -> Portgroup:
        portgroup = Portgroup(self, group_id, self._next_portgroup_id,
                              port_mode, ports)
        self._next_portgroup_id += 1
        return portgroup

    def port_type_shown(self, full_port_type: tuple[PortType, PortSubType]) -> bool:
        port_type, sub_type = full_port_type
        if port_type is PortType.MIDI_JACK:
            return bool(self.port_types_view & PortTypesViewFlag.MIDI)
        if port_type is PortType.AUDIO_JACK:
            if sub_type is PortSubType.REGULAR:
                return bool(self.port_types_view & PortTypesViewFlag.AUDIO)
            elif sub_type is PortSubType.CV:
                return bool(self.port_types_view & PortTypesViewFlag.CV)
            elif sub_type is (PortSubType.REGULAR | PortSubType.CV):
                return bool(self.port_types_view & (PortTypesViewFlag.AUDIO
                                                    | PortTypesViewFlag.CV))
        if port_type is PortType.MIDI_ALSA:
            return bool(self.alsa_midi_enabled
                        and self.port_types_view & PortTypesViewFlag.ALSA)
        if port_type is PortType.VIDEO:
            return bool(self.port_types_view & PortTypesViewFlag.VIDEO)

        return False

    def get_group_from_name(self, group_name: str) -> Union[Group, None]:
        return self._groups_by_name.get(group_name)

    def get_group_from_id(self, group_id: int) -> Union[Group, None]:
        return self._groups_by_id.get(group_id)

    def get_port_from_name(self, port_name: str) -> Port:
        return self._ports_by_name.get(port_name)

    def get_port_from_uuid(self, uuid:int) -> Port:
        for group in self.groups:
            for port in group.ports:
                if port.uuid == uuid:
                    return port

    def get_port_from_id(self, group_id: int, port_id: int) -> Port:
        group = self.get_group_from_id(group_id)
        if group is not None:
            for port in group.ports:
                if port.port_id == port_id:
                    return port

    def save_group_position(self, gpos: GroupPos):
        # reimplement this to save a group position elsewhere
        pass

    def save_portgroup_memory(self, portgrp_mem: PortgroupMem):
        # reimplement this to save a portgroup memory elsewhere
        pass

    def get_corrected_a2j_group_name(self, group_name: str) -> str:
        # a2j replace points with spaces in the group name
        # we do nothing here, but in some conditions we can 
        # assume we know it.
        return group_name

    def set_group_as_nsm_client(self, group: Group):
        pass

    def get_group_position(self, group_name: str) -> GroupPos:
        for gpos in self.group_positions:
            if (gpos.port_types_view is self.port_types_view
                    and gpos.group_name == group_name):
                return gpos

        # prevent move to a new position in case of port_types_view change
        # if there is no remembered position for this group in new view
        group = self.get_group_from_name(group_name)
        if group is not None:
            # copy the group_position
            gpos = group.current_position.copy()
            gpos.port_types_view = self.port_types_view
            self.group_positions.append(gpos)
            self.save_group_position(gpos)
            return gpos

        # group position doesn't already exists, create one
        gpos = GroupPos()
        gpos.fully_set = False
        gpos.port_types_view = self.port_types_view
        gpos.group_name = group_name
        gpos.null_xy, gpos.in_xy, gpos.out_xy = get_new_group_positions()
        self.group_positions.append(gpos)
        self.save_group_position(gpos)
        return gpos

    def add_portgroup_memory(self, portgroup_mem: PortgroupMem):
        remove_list = list[PortgroupMem]()

        for pg_mem in self.portgroups_memory:
            if pg_mem.has_a_common_port_with(portgroup_mem):
                remove_list.append(pg_mem)

        for pg_mem in remove_list:
            self.portgroups_memory.remove(pg_mem)

        self.portgroups_memory.append(portgroup_mem)
        
        group = self.get_group_from_name(portgroup_mem.group_name)
        if group is not None:
            group.portgroup_memory_added(portgroup_mem)

    def remove_and_add_all(self):
        self.optimize_operation(True)
            
        for connection in self.connections:
            connection.remove_from_canvas()
        
        for group in self.groups:
            for portgroup in group.portgroups:
                portgroup.remove_from_canvas()
            
            for port in group.ports:
                port.remove_from_canvas()
            group.remove_from_canvas()
            
            group.add_to_canvas()
            for port in group.ports:
                port.add_to_canvas()
            for portgroup in group.portgroups:
                portgroup.add_to_canvas()
        
        for connection in self.connections:
            connection.add_to_canvas()
        
        self.optimize_operation(False)
        patchcanvas.redraw_all_groups()

    def clear_all(self):
        self.optimize_operation(True)
        for connection in self.connections:
            connection.remove_from_canvas()

        for group in self.groups:
            group.remove_all_ports()
            group.remove_from_canvas()

        self.optimize_operation(False)

        self.connections.clear()
        self._clear_groups()

        patchcanvas.clear()

        self._next_group_id = 0
        self._next_port_id = 0
        self._next_portgroup_id = 1
        self._next_connection_id = 0

    def change_port_types_view(self, port_types_view: PortTypesViewFlag, force=False):
        if not force and port_types_view is self.port_types_view:
            return
        
        self.port_types_view = port_types_view

        # Prevent visual update at each canvas item creation
        # because we may create/remove a lot of ports here
        self.optimize_operation(True)

        for connection in self.connections:
            if (connection.in_canvas
                    and not connection.shown_in_port_types_view(port_types_view)):
                connection.remove_from_canvas()

        groups_and_pos = dict[Group, GroupPos]()

        for group in self.groups:
            group.change_port_types_view()
            groups_and_pos[group] = self.get_group_position(group.name)

        for connection in self.connections:
            connection.add_to_canvas()

        self.optimize_operation(False)
        
        patchcanvas.redraw_all_groups(force_no_prevent_overlap=True)

        for group, gpos in groups_and_pos.items():
            group.set_group_position(gpos, view_change=True)

        patchcanvas.repulse_all_boxes()
        self.sg.port_types_view_changed.emit(self.port_types_view)

    # --- options triggers ---

    def set_graceful_names(self, yesno: int):
        if self.use_graceful_names != yesno:
            self.toggle_graceful_names()

    def set_a2j_grouped(self, yesno: int):
        if self.group_a2j_hw != bool(yesno):
            self.group_a2j_hw = bool(yesno)
            self.refresh()

    def set_alsa_midi_enabled(self, yesno: int):
        if self.alsa_midi_enabled != bool(yesno):
            self.alsa_midi_enabled = bool(yesno)
            self.change_port_types_view(PortTypesViewFlag.ALL, force=True)

    def set_group_shadows(self, yesno: int):
        patchcanvas.options.show_shadows = bool(yesno)
        self.remove_and_add_all()

    def set_auto_select_items(self, yesno: int):
        patchcanvas.set_auto_select_items(bool(yesno))

    def change_theme(self, theme_name: str):
        if not theme_name:
            return
        patchcanvas.change_theme(theme_name)

    def set_elastic_canvas(self, yesno: int):
        patchcanvas.set_elastic(bool(yesno))

    def set_borders_navigation(self, yesno: int):
        patchcanvas.set_borders_navigation(bool(yesno))

    def set_prevent_overlap(self, yesno: int):
        patchcanvas.set_prevent_overlap(bool(yesno))

    def toggle_graceful_names(self):
        self.set_use_graceful_names(not self.use_graceful_names)
        self.optimize_operation(True)
        for group in self.groups:
            group.update_ports_in_canvas()
            group.update_name_in_canvas()
        self.optimize_operation(False)
        patchcanvas.redraw_all_groups()

    def set_zoom(self, zoom: float):
        patchcanvas.canvas.scene.zoom_ratio(zoom)

    def zoom_reset(self):
        patchcanvas.zoom_reset()
    
    def zoom_fit(self):
        patchcanvas.zoom_fit()  

    def refresh(self):
        self.clear_all()

    @later_by_batch()
    def set_group_uuid_from_name(self, client_name: str, uuid: int):
        group = self.get_group_from_name(client_name)
        if group is not None:
            group.uuid = uuid

    @later_by_batch(draw_group=True)
    def add_port(self, name: str, port_type_int: int, flags: int, uuid: int) -> int:
        ''' adds port and returns the group_id '''
        port_type = PortType.NULL

        if port_type_int == PortType.AUDIO_JACK:
            port_type = PortType.AUDIO_JACK
        elif port_type_int == PortType.MIDI_JACK:
            port_type = PortType.MIDI_JACK
        elif port_type_int == PortType.MIDI_ALSA:
            port_type = PortType.MIDI_ALSA
        elif port_type_int == PortType.VIDEO:
            port_type = PortType.VIDEO

        port = Port(self, self._next_port_id, name, port_type, flags, uuid)
        self._next_port_id += 1
        
        full_port_name = name
        group_name, colon, port_name = full_port_name.partition(':')

        is_a2j_group = False
        group_is_new = False

        if (port_type is PortType.MIDI_ALSA
                and full_port_name.startswith((':ALSA_OUT:', ':ALSA_IN:'))):
            _, _alsa_key, alsa_gp_id, alsa_p_id, group_name, *rest = \
                full_port_name.split(':')
            port_name = ':'.join(rest)

            if port.flags & JackPortFlag.IS_PHYSICAL:
                is_a2j_group = True

        elif (full_port_name.startswith(('a2j:', 'Midi-Bridge:'))
                and (not self.group_a2j_hw
                     or not port.flags & JackPortFlag.IS_PHYSICAL)):
            group_name, colon, port_name = port_name.partition(':')
            if full_port_name.startswith('a2j:'):
                if ' [' in group_name:
                    group_name = group_name.rpartition(' [')[0]
                else:
                    if ' (capture)' in group_name:
                        group_name = group_name.partition(' (capture)')[0]
                    else:
                        group_name = group_name.partition(' (playback)')[0]

                # fix a2j wrongly substitute '.' with space
                group_name = self.get_corrected_a2j_group_name(group_name)

            if port.flags & JackPortFlag.IS_PHYSICAL:
                is_a2j_group = True
        
        group = self.get_group_from_name(group_name)
        if group is None:
            # port is in an non existing group, create the group
            gpos = self.get_group_position(group_name)
            group = Group(self, self._next_group_id, group_name, gpos)
            group.a2j_group = is_a2j_group
            self.set_group_as_nsm_client(group)

            self._next_group_id += 1
            self._add_group(group)
            group_is_new = True

        group.add_port(port)
        group.graceful_port(port)

        if group_is_new:
            gpos = self.get_group_position(group_name)
            group.set_group_position(gpos)

        group.add_to_canvas()
        port.add_to_canvas()
        group.check_for_portgroup_on_last_port()
        group.check_for_display_name_on_last_port()
        
        return group.group_id

    @later_by_batch(draw_group=True, clear_conns=True)
    def remove_port(self, name: str) -> Union[int, None]:
        ''' removes a port from name and return its group_id '''
        port = self.get_port_from_name(name)
        if port is None:
            return None

        group = self.get_group_from_id(port.group_id)
        if group is None:
            return None

        # remove portgroup first if port is in a portgroup
        if port.portgroup_id:
            for portgroup in group.portgroups:
                if portgroup.portgroup_id == port.portgroup_id:
                    group.portgroups.remove(portgroup)
                    portgroup.remove_from_canvas()
                    break

        port.remove_from_canvas()
        group.remove_port(port)

        if not group.ports:
            group.remove_from_canvas()
            self._remove_group(group)
            return None
        
        return group.group_id

    @later_by_batch(draw_group=True)
    def rename_port(self, name: str, new_name: str) -> Union[int, None]:
        port = self.get_port_from_name(name)
        if port is None:
            _logger.warning(f"rename_port to '{new_name}', no port named '{name}'")
            return

        # change port key in self._ports_by_name dict
        self._ports_by_name.pop(name)
        self._ports_by_name[new_name] = port

        group_name = name.partition(':')[0]
        new_group_name = new_name.partition(':')[0]

        # In case a port rename implies another group for the port
        if group_name != new_group_name:
            group = self.get_group_from_name(group_name)
            if group is not None:
                group.remove_port(port)
                if not group.ports:
                    self._remove_group(group)

            port.remove_from_canvas()
            port.full_name = new_name

            group = self.get_group_from_name(new_group_name)
            if group is None:
                # copy the group_position to not move the group
                # because group has been renamed
                orig_gpos = self.get_group_position(group_name)
                gpos = orig_gpos.copy()
                gpos.group_name = new_group_name

                group = Group(self, self._next_group_id, new_group_name, gpos)
                self._next_group_id += 1
                group.add_port(port)
                group.add_to_canvas()
            else:
                group.add_port(port)

            port.add_to_canvas()
            return group.group_id

        group = self.get_group_from_id(port.group_id)
        if group is not None:
            # because many ports may be renamed quicky
            # It is prefferable to rename all theses ports together.
            # It prevents too much widget update in canvas,
            # renames now could also prevent to find stereo detected portgroups
            # if one of the two ports has been renamed and not the other one.
            port.full_name = new_name
            group.graceful_port(port)
            port.rename_in_canvas()

            return group.group_id

    @later_by_batch(sort_group=True)
    def metadata_update(self, uuid: int, key: str, value: str) -> Optional[int]:
        ''' remember metadata and returns the group_id'''
        if key == JACK_METADATA_ORDER:
            port = self.get_port_from_uuid(uuid)
            if port is None:
                return

            try:
                port_order = int(value)
            except:
                _logger.warning(
                    f"JACK_METADATA_ORDER for UUID {uuid} "
                    f"value '{value}' is not an int")
                return

            port.order = port_order
            return port.group_id

        elif key == JACK_METADATA_PRETTY_NAME:
            port = self.get_port_from_uuid(uuid)
            if port is None:
                return

            port.pretty_name = value
            port.rename_in_canvas()
            return port.group_id

        elif key == JACK_METADATA_PORT_GROUP:
            port = self.get_port_from_uuid(uuid)
            if port is None:
                return

            port.mdata_portgroup = value
            return port.group_id

        elif key == JACK_METADATA_ICON_NAME:
            for group in self.groups:
                if group.uuid == uuid:
                    group.set_client_icon(value, from_metadata=True)
                    return group.group_id

        elif key == JACK_METADATA_SIGNAL_TYPE:
            port = self.get_port_from_uuid(uuid)
            if port is None:
                return

            port.mdata_signal_type = value

            if port.type is PortType.AUDIO_JACK:
                if value == 'CV':
                    port.subtype = PortSubType.CV
                elif value == 'AUDIO':
                    port.subtype = PortSubType.REGULAR

            return port.group_id

    @later_by_batch()
    def add_connection(self, port_out_name: str, port_in_name: str):
        port_out = self.get_port_from_name(port_out_name)
        port_in = self.get_port_from_name(port_in_name)  

        if port_out is None or port_in is None:
            return

        for connection in self.connections:
            if (connection.port_out == port_out
                    and connection.port_in == port_in):
                return

        connection = Connection(self, self._next_connection_id, port_out, port_in)
        self._next_connection_id += 1
        self.connections.append(connection)
        connection.add_to_canvas()

    @later_by_batch()
    def remove_connection(self, port_out_name: str, port_in_name: str):
        port_out = self.get_port_from_name(port_out_name)
        port_in = self.get_port_from_name(port_in_name)
        if port_out is None or port_in is None:
            return

        for connection in self.connections:
            if (connection.port_out == port_out
                    and connection.port_in == port_in):
                self.connections.remove(connection)
                connection.remove_from_canvas()
                break

    def disannounce(self):
        self.clear_all()

    @in_main_thread()
    def server_started(self):
        self.server_is_started = True
        if self._tools_widget is not None:
            self._tools_widget.set_jack_running(True)

    @in_main_thread()
    def server_stopped(self):
        self.server_is_started = False
        if self._tools_widget is not None:
            self._tools_widget.set_jack_running(
                False, use_alsa_midi=self.alsa_midi_enabled)

        self.clear_all()
        if self.alsa_midi_enabled:
            self.refresh()

    @in_main_thread()
    def server_lose(self):
        self.server_is_started = False
        
        if self._tools_widget is not None:
            self._tools_widget.set_jack_running(False)

        self.clear_all()
        if self.alsa_midi_enabled:
            self.refresh()

        if self.main_win is not None:
            ret = QMessageBox.critical(
                self.main_win,
                _translate('patchbay', "JACK server lose"),
                _translate('patchbay', "JACK server seems to be totally busy... ;("))

    @in_main_thread()
    def set_dsp_load(self, dsp_load: int):
        if self._tools_widget is not None:
            self._tools_widget.set_dsp_load(dsp_load)

    @in_main_thread()
    def add_xrun(self):
        if self._tools_widget is not None:
            self._tools_widget.add_xrun()

    @in_main_thread()
    def refresh_transport(self, transport_position: TransportPosition):
        if self._tools_widget is not None:
            self._tools_widget.refresh_transport(transport_position)

    def change_buffersize(self, buffer_size: int):
        pass

    def transport_play_pause(self, play: bool):
        pass

    def transport_stop(self):
        pass

    def transport_relocate(self, frame: int):
        pass

    def change_tools_displayed(self, tools_displayed: ToolDisplayed):
        if self._tools_widget is not None:
            self._tools_widget.change_tools_displayed(tools_displayed)

        self._tools_displayed = tools_displayed

    def redraw_all_groups(self):
        patchcanvas.redraw_all_groups()

    def filter_groups(self, text: str, n_select=0) -> int:
        ''' semi hides groups not matching with text
            and returns number of matching boxes '''
        opac_grp_ids = set()
        opac_conn_ids = set()

        for group in self.groups:
            opac = bool(text.lower() not in group.name.lower()
                        and text.lower() not in group.display_name.lower())
            if opac:
                opac_grp_ids.add(group.group_id)

            group.semi_hide(opac)
        
        for conn in self.connections:
            opac_conn = bool(
                conn.port_out.group_id in opac_grp_ids
                and conn.port_in.group_id in opac_grp_ids)
            
            conn.semi_hide(opac_conn)
            if opac_conn:
                opac_conn_ids.add(conn.connection_id)
            
        for group in self.groups:
            if group.group_id in opac_grp_ids:
                group.set_in_front()
        
        for conn in self.connections:
            if conn.connection_id in opac_conn_ids:
                conn.set_in_front()
        
        for conn in self.connections:
            if conn.connection_id not in opac_conn_ids:
                conn.set_in_front()
        
        n_boxes = 0
        
        for group in self.groups:
            if group.group_id not in opac_grp_ids:
                group.set_in_front()
                n_grp_boxes = group.get_number_of_boxes()

                if n_select > n_boxes and n_select <= n_boxes + n_grp_boxes:
                    group.select_filtered_box(n_select - n_boxes)
                n_boxes += n_grp_boxes

        return n_boxes

    def set_semi_hide_opacity(self, opacity: float):
        patchcanvas.set_semi_hide_opacity(opacity)

    @in_main_thread()
    def buffer_size_changed(self, buffer_size: int):
        if self._tools_widget is not None:
            self._tools_widget.set_buffer_size(buffer_size)

    @in_main_thread()
    def sample_rate_changed(self, samplerate: int):
        if self._tools_widget is not None:
            self._tools_widget.set_samplerate(samplerate)

    def _delayed_orders_timeout(self):
        self.optimize_operation(True)
        group_ids_to_update = set()
        group_ids_to_sort = set()
        some_groups_removed = False
        clear_conns = False
        
        for oq in self.delayed_orders:
            group_id = oq.func(*oq.args, **oq.kwargs)
            if oq.sort_group and group_id is not None:
                group_ids_to_sort.add(group_id)
            if oq.draw_group:
                if group_id is not None:
                    group_ids_to_update.add(group_id)
                else:
                    some_groups_removed = True
            if oq.clear_conns:
                clear_conns = True
        
        for group in self.groups:
            if group.group_id in group_ids_to_sort:
                group.sort_ports_in_canvas()

        if clear_conns:
            # sometimes connections are still in canvas without ports
            # probably because the message for their destruction
            # has not been received.
            # here we can assume to clear all of them
            conns_to_clean = list[Connection]()

            for conn in self.connections:
                for port in (conn.port_out, conn.port_in):
                    fport = self.get_port_from_name(port.full_name)
                    if fport is None:
                        conn.remove_from_canvas()
                        conns_to_clean.append(conn)
                        break
                    
                    if not fport.in_canvas:
                        conn.remove_from_canvas()
                    
            for conn in conns_to_clean:
                self.connections.remove(conn)
        
        self.optimize_operation(False)
        self.delayed_orders.clear()

        for group in self.groups:
            if group.group_id in group_ids_to_update:
                group.redraw_in_canvas()

        if some_groups_removed:
            patchcanvas.canvas.scene.resize_the_scene()

    def _export_port_list_to_patchichi(self) -> str:
        def slcol(input_str: str) -> str:
            return input_str.replace(':', '\\:')
        
        contents = ''

        gps_and_ports = list[tuple[str, list[Port]]]()
        for group in self.groups:
            for port in group.ports:
                if port.type is PortType.MIDI_ALSA:
                    group_name = port.full_name.split(':')[4]
                    print('goreupp name', group_name)
                else:
                    group_name = port.full_name.partition(':')[0]

                for gp_name, gp_port_list in gps_and_ports:
                    if gp_name == group_name:
                        gp_port_list.append(port)
                        break
                else:
                    gps_and_ports.append((group_name, [port]))

        for group_name, port_list in gps_and_ports:
            port_list.sort(key=operator.attrgetter('port_id'))
                
        for group_name, port_list in gps_and_ports:
            gp_written = False              
            last_type_and_mode = (PortType.NULL, PortMode.NULL)
            physical = False
            terminal = False
            monitor = False
            pg_name = ''
            signal_type = ''

            for port in port_list:
                if not gp_written:
                    contents += f'\n::{group_name}\n'
                    gp_written = True

                    group = self.get_group_from_name(group_name)
                    if group is not None:
                        group_attrs = list[str]()
                        if group.client_icon:
                            group_attrs.append(f'CLIENT_ICON={slcol(group.client_icon)}')
                            
                        if group.mdata_icon:
                            group_attrs.append(f'ICON_NAME={slcol(group.mdata_icon)}')

                        if group.has_gui:
                            if group.gui_visible:
                                group_attrs.append('GUI_VISIBLE')
                            else:
                                group_attrs.append('GUI_HIDDEN')
                        if group_attrs:
                            contents += ':'
                            contents += '\n:'.join(group_attrs)
                            contents += '\n'

                if port.flags & JackPortFlag.IS_PHYSICAL:
                    if not physical:
                        contents += ':PHYSICAL\n'
                        physical = True
                elif physical:
                    contents += ':~PHYSICAL\n'
                    physical = False

                if last_type_and_mode != (port.type, port.mode()):
                    if port.type is PortType.AUDIO_JACK:
                        if port.flags & JackPortFlag.IS_CONTROL_VOLTAGE:
                            contents += ':CV'
                        else:
                            contents += ':AUDIO'
                    elif port.type is PortType.MIDI_JACK:
                        contents += ':MIDI'
                    elif port.type is PortType.MIDI_ALSA:
                        contents += ':ALSA'

                    contents += f':{port.mode().name}\n'
                    last_type_and_mode = (port.type, port.mode())
                
                if port.mdata_signal_type != signal_type:
                    if port.mdata_signal_type:
                        contents += f':SIGNAL_TYPE={slcol(port.mdata_signal_type)}\n'
                    else:
                        contents += ':~SIGNAL_TYPE\n'
                
                if port.mdata_portgroup != pg_name:
                    if port.mdata_portgroup:
                        contents += f':PORTGROUP={slcol(port.mdata_portgroup)}\n'
                    else:
                        contents += ':~PORTGROUP\n'
                    pg_name = port.mdata_portgroup

                if port.type is PortType.MIDI_ALSA:
                    port_short_name = ':'.join(port.full_name.split(':')[5:])
                else:
                    port_short_name = port.full_name.partition(':')[2]

                contents += f'{port_short_name}\n'
                
                if port.pretty_name or port.order:
                    port_attrs = list[str]()
                    if port.pretty_name:
                        port_attrs.append(f'PRETTY_NAME={slcol(port.pretty_name)}')
                    if port.order:
                        port_attrs.append(f'ORDER={port.order}')
                    contents += ':'
                    contents += '\n:'.join(port_attrs)
                    contents += '\n'

        return contents
    
    def export_to_patchichi_json(
            self, path: Path, editor_text='') -> bool:
        if not editor_text:
            editor_text = self._export_port_list_to_patchichi()

        file_dict = dict[str, Any]()
        file_dict['VERSION'] = (0, 1)       
        file_dict['editor_text'] = editor_text
        file_dict['connections'] = [
            (c.port_out.full_name, c.port_in.full_name)
            for c in self.connections]
        file_dict['group_positions'] = [
            gpos.as_serializable_dict() for gpos in self.group_positions
            if self.get_group_from_name(gpos.group_name) is not None]
        
        portgroups = list[dict[str, Any]]()
        for pg_mem in self.portgroups_memory:
            group = self.get_group_from_name(pg_mem.group_name)
            if group is None:
                continue

            for port_str in pg_mem.port_names:
                for port in group.ports:
                    if (port.short_name() == port_str
                            and port.type == pg_mem.port_type
                            and port.mode() == pg_mem.port_mode):
                        portgroups.append(pg_mem.as_serializable_dict())
                        break
        
        file_dict['portgroups'] = portgroups
        
        try:
            with open(path, 'w') as f:
                json.dump(file_dict, f, indent=2)
            return True
        except Exception as e:
            _logger.error(f'Failed to save patchichi file: {str(e)}')
            return False
        
            
        


from enum import IntEnum, IntFlag

class JackOptions(IntFlag):
    NULL = 0x00
    NO_START_SERVER = 0x01
    USE_EXACT_NAME = 0x02
    SERVER_NAME = 0x04
    LOAD_NAME = 0x08
    LOAD_INIT = 0x10
    SESSION_ID = 0x20
    
    OPEN_OPTIONS = SESSION_ID | SERVER_NAME | NO_START_SERVER | USE_EXACT_NAME
    LOAD_OPTIONS = LOAD_INIT | LOAD_NAME | USE_EXACT_NAME


class JackStatus(IntFlag):
    FAILURE = 0x01
    INVALID_OPTION = 0x02
    NAME_NOT_UNIQUE = 0x04
    SERVER_STARTED = 0x08
    SERVER_FAILED = 0x10
    SERVER_ERROR = 0x20
    NO_SUCH_CLIENT = 0x40
    LOAD_FAILURE = 0x80
    INIT_FAILURE = 0x100
    SHM_FAILURE = 0x200
    VERSION_ERROR = 0x400
    BACKEND_ERROR = 0x800
    CLIENT_ZOMBIE = 0x1000


class JackLatencyCallbackMode(IntEnum):
    CAPTURE = 0
    PLAYBACK = 1


class JackPortFlags(IntFlag):
    IS_INPUT = 0x01
    IS_OUTPUT = 0x02
    IS_PHYSICAL = 0x04
    CAN_MONITOR = 0x08
    IS_TERMINAL = 0x10
    IS_CONTROL_VOLTAGE = 0x100


class JackTransportState(IntEnum):
    STOPPED = 0
    ROLLING = 1
    LOOPING = 2
    STARTING = 3
    # JACK2 only:
    NET_STARTING = 4
    

class JackPositionBits(IntFlag):
    POSITION_BBT = 0x10
    POSITION_TIMECODE = 0x20
    BBT_FRAME_OFFSET = 0x40
    AUDIO_VIDEO_RATIO = 0x80
    VIDEO_FRAME_OFFSET = 0x100
    
    POSITION_MASK = (
        POSITION_BBT
        | POSITION_TIMECODE
        | BBT_FRAME_OFFSET
        | AUDIO_VIDEO_RATIO
        | VIDEO_FRAME_OFFSET)


class JackSessionEvenType(IntEnum):
    SAVE = 1
    SAVE_AND_QUIT = 2
    SAVE_TEMPLATE = 3


class JackSessionFlags(IntFlag):
    SAVE_ERROR = 0x01
    NEED_TERMINAL = 0x02


class JackPropertyChange(IntEnum):
    CREATED = 0
    CHANGED = 1
    DELETED = 2
    

class JackMetadata:
    _PREFIX = "http://jackaudio.org/metadata/"
    CONNECTED = _PREFIX + "connected"
    EVENT_TYPES = _PREFIX + "event-types"
    HARDWARE = _PREFIX + "hardware"
    ICON_LARGE = _PREFIX + "icon-large"
    ICON_NAME = _PREFIX + "icon-name"
    ICON_SMALL = _PREFIX + "icon-small"
    ORDER = _PREFIX + "order"
    PORT_GROUP = _PREFIX + "port-group"
    PRETTY_NAME = _PREFIX + "pretty-name"
    SIGNAL_TYPE = _PREFIX + "signal-type"



from typing import TYPE_CHECKING, Union
from PyQt5.QtCore import Qt, QSize
from PyQt5.QtWidgets import (QCheckBox, QFrame, QHBoxLayout, QLabel, 
                             QSpacerItem, QSizePolicy)
from PyQt5.QtGui import QPixmap, QColor


from .utils import get_group_icon
from .theme import StyleAttributer
from .init_values import (
    GroupObject,
    BoxType,
    PortObject,
    PortSubType,
    PortgrpObject,
    canvas,
    PortType)

if TYPE_CHECKING:
    from .connect_menu import SubMenu


def theme_css(theme: StyleAttributer) -> str:
    pen = theme.fill_pen()
    
    return (f"background-color: {theme.background_color().name(QColor.HexArgb)};"
            f"color: {theme.text_color().name(QColor.HexArgb)};"
            f"border: {pen.widthF()}px solid {pen.color().name(QColor.HexArgb)}")

    
class PortCheckBox(QCheckBox):
    def __init__(self, p_object:  Union[PortObject, PortgrpObject],
                 parent: 'SubMenu'):
        QCheckBox.__init__(self, "", parent)
        self.setTristate(True)
        self._p_object = p_object
        self._parent = parent
        self.set_theme()

    def set_theme(self):        
        po = self._p_object
        
        theme = canvas.theme.port
        line_theme = canvas.theme.line
        
        if isinstance(po, PortgrpObject):
            theme = canvas.theme.portgroup
            
        if po.port_type is PortType.AUDIO_JACK:
            theme = theme.audio
            line_theme = line_theme.audio
        elif po.port_type is PortType.MIDI_JACK:
            theme = theme.midi
            line_theme = line_theme.midi
        elif po.port_type is PortType.MIDI_ALSA:
            theme = theme.alsa
            line_theme = line_theme.alsa

        text_color = theme.text_color().name(QColor.HexArgb)
        border_color = theme.fill_pen().color().name(QColor.HexArgb)
        h_text_color = theme.selected.text_color().name(QColor.HexArgb)
        ind_bg = canvas.theme.scene_background_color.name(QColor.HexArgb)
        checked_bg = line_theme.selected.background_color().name(QColor.HexArgb)
        
        border_width = theme.fill_pen().widthF()
        
        TOP, RIGHT, BOTTOM, LEFT = 0, 1, 2, 3
        SIDES = ['top', 'right', 'bottom', 'left']
        margin_texts = [f"margin-{side}: 2px" for side in SIDES]
        border_texts = [f"border-{side}: {border_width}px solid {border_color}"
                        for side in SIDES]
        radius_text = ""
        
        if isinstance(po, PortObject) and po.portgrp_id:
            if po.pg_pos == 0:
                margin_texts.pop(BOTTOM)
                radius_text = "border-bottom-left-radius: 0px; border-bottom-right-radius: 0px"
            elif po.pg_pos + 1 == po.pg_len:
                margin_texts.pop(TOP)
                radius_text = "border-top-left-radius: 0px; border-top-right-radius: 0px"
                
            if po.pg_pos != 0:
                border_texts[TOP] = f"border-top: 0px solid transparent"


        self.setStyleSheet(
            f"QCheckBox{{background-color: none;color: {text_color}; spacing: 0px;"
                       f"border-radius: 3px; {radius_text};}}"
            f"QCheckBox:hover{{background-color: none;color: {h_text_color}}}"
        )
        # # this stylesheet code is commented because the checkbox was not clear 
        # # with some Qt theme (fusion, kvantum).
        # # seems impossible to have a coherent and readable style
        # # because change indicator size seems to not be applied when box is checked
        # # and it is very ugly.
        # # So we keep native checkboxes for the moment to ensure readability,
        # # and will see later if we found a better solution.

        #     f"QCheckBox::indicator{{width: 12px;height: 12px;"
        #                           f"background-color: {ind_bg};margin: 3px;"
        #                           f"border-radius: 3px; border: 1px solid "
        #                           f"{theme.fill_pen().color().name()}}}"
        #     f"QCheckBox::indicator:checked{{"
        #         f"background-color: {checked_bg}; border: 3px solid {ind_bg}}}"
        #     f"QCheckBox::indicator:indeterminate{{width: 12px;height: 12px;"
        #         f"background-color: qlineargradient("
        #             f"x1: 0, y1: 0, x2: 1, y2: 1, "
        #             f"stop: 0 {checked_bg}, stop: 0.55 {checked_bg}, "
        #             f"stop: 0.60 {ind_bg}, stop: 1 {ind_bg}); "
        #         f"border: 3px solid {ind_bg}}}")

    def nextCheckState(self):
        po = self._p_object
        port_id = po.port_id if isinstance(po, PortObject) else -1
        
        self._parent.connection_asked_from_box(po, not self.isChecked())


class CheckFrame(QFrame):
    def __init__(self, p_object: Union[PortObject, PortgrpObject],
                 port_name: str, port_name_end: str,
                 parent: 'SubMenu'):
        QFrame.__init__(self, parent)
        # self.setMinimumSize(QSize(100, 18))
        
        self._p_object = p_object
        self._parent = parent
        
        self._check_box = PortCheckBox(p_object, parent)
        self._label_left = QLabel(port_name)
        self._layout = QHBoxLayout(self)
        self._layout.setSpacing(0)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.addWidget(self._check_box)
        self._layout.addWidget(self._label_left)
        spacer = QSpacerItem(2, 2, QSizePolicy.Expanding, QSizePolicy.Minimum)
        self._layout.addSpacerItem(spacer)
        self._label_right = None
        if port_name_end:
            self._label_right = QLabel(port_name_end)
            self._layout.addWidget(self._label_right)
        self._set_theme()

    def _set_theme(self):
        po = self._p_object

        theme = canvas.theme.port
        line_theme = canvas.theme.line        

        if isinstance(po, PortgrpObject):
            theme = canvas.theme.portgroup
            
        if po.port_type is PortType.AUDIO_JACK:
            theme = theme.cv if po.port_subtype is PortSubType.CV else theme.audio
            line_theme = line_theme.audio
            
        elif po.port_type is PortType.MIDI_JACK:
            theme = theme.midi
            line_theme = line_theme.midi
            
        elif po.port_type is PortType.MIDI_ALSA:
            theme = theme.alsa
            line_theme = line_theme.alsa

        text_color = theme.text_color().name()
        border_color = theme.fill_pen().color().name()
        h_text_color = theme.selected.text_color().name()
        
        border_width = theme.fill_pen().widthF()
        
        TOP, RIGHT, BOTTOM, LEFT = 0, 1, 2, 3
        SIDES = ['top', 'right', 'bottom', 'left']
        margin_texts = [f"margin-{side}: 2px" for side in SIDES]
        border_texts = [f"border-{side}: {border_width}px solid {border_color}"
                        for side in SIDES]
        radius_text = ""
        
        if isinstance(po, PortObject) and po.portgrp_id:
            if po.pg_pos == 0:
                margin_texts.pop(BOTTOM)
                radius_text = "border-bottom-left-radius: 0px; border-bottom-right-radius: 0px"
            elif po.pg_pos + 1 == po.pg_len:
                margin_texts.pop(TOP)
                radius_text = "border-top-left-radius: 0px; border-top-right-radius: 0px"
                
            if po.pg_pos != 0:
                border_texts[TOP] = f"border-top: 0px solid transparent"

        margins_text = ';'.join(margin_texts)
        borders_text = ';'.join(border_texts)

        self.setFont(theme.font())
        self.setStyleSheet(
            f"CheckFrame{{{theme_css(theme)}; spacing: 0px;"
            f"{borders_text}; border-radius: 3px; {radius_text};"
            f"{margins_text}; padding-right: 0px}}"
            f"CheckFrame:focus{{{theme_css(theme.selected)}}};")
        
        self._label_left.setFont(theme.font())
        self._label_left.setStyleSheet(
            f"QLabel{{color: {text_color}}} QLabel:focus{{color: {h_text_color}}} ")
        
        if self._label_right is not None:
            port_theme = canvas.theme.port
            if po.port_type is PortType.AUDIO_JACK:
                port_theme = port_theme.audio
            elif po.port_type is PortType.MIDI_JACK:
                port_theme = port_theme.midi

            self._label_right.setFont(port_theme.font())
            self._label_right.setStyleSheet(
                f"QLabel{{margin-left: 3px; margin-right: 0px; padding: 0px; {theme_css(port_theme)}}} "
                f"QLabel:focus{{{theme_css(port_theme.selected)}}}")

    def set_check_state(self, check_state: int):
        self._check_box.setCheckState(check_state)

    def connection_asked_from_box(self, group_id: int, port_id: int,
                                  portgrp_id: int, yesno: bool):
        self._parent.connection_asked_from_box(group_id, port_id, portgrp_id, yesno)
    
    def mousePressEvent(self, event):
        self._check_box.nextCheckState()
        
    def keyPressEvent(self, event):
        if event.key() in (Qt.Key_Space, Qt.Key_Return):
            self._check_box.nextCheckState()
            return
        QFrame.keyPressEvent(self, event)
        
    def enterEvent(self, event):
        super().enterEvent(event)
        self.setFocus()
        # self._label_right.setSelec
        

class GroupFrame(QFrame):
    ''' this class is not used now. The idea was to have group menu actions
        styled as their boxes are in the canvas, but it implies too much
        changes. Default Qt theme works for QMenu works better for the moment. '''

    def __init__(self, group: GroupObject, parent: 'SubMenu'):
        super().__init__(parent)
        self._group = group
        
        icon = get_group_icon(group.group_id, parent._port_mode)
        self._label_icon = QLabel()
        self._label_icon.setPixmap(icon.pixmap(QSize(16, 16)))
        self._label_group_name = QLabel(group.group_name)

        self._layout = QHBoxLayout(self)
        self._layout.setSpacing(0)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.addWidget(self._label_icon)
        self._layout.addWidget(self._label_group_name)
        
        theme = canvas.theme.box
        self._label_group_name.setStyleSheet(f"GroupFrame{{{theme_css(theme)}}}")
        



import os
import shutil
import time

from PyQt5.QtWidgets import (QApplication, QTreeWidget, QTreeWidgetItem,
                             QDialogButtonBox, QMenu, QMessageBox)
from PyQt5.QtGui import QIcon, QCursor
from PyQt5.QtCore import Qt, QTimer, QDateTime, QLocale, QPoint
from typing import Optional

import ray
import child_dialogs
from gui_tools import CommandLineArgs, RayIcon, is_dark_theme, basename
from child_dialogs import ChildDialog
from client_properties_dialog import ClientPropertiesDialog
from snapshots_dialog import (
    Snapshot, SnapGroup, GROUP_MAIN)

import ui.open_session

_translate = QApplication.translate

COLUMN_NAME = 0
COLUMN_NOTES = 1
COLUMN_SCRIPTS = 2
COLUMN_DATE = 3

PENDING_ACTION_NONE = 0
PENDING_ACTION_RENAME = 1
PENDING_ACTION_DUPLICATE = 2
PENDING_ACTION_TEMPLATE = 3

CORNER_HIDDEN = 0
CORNER_LISTING = 1
CORNER_COPY = 2
CORNER_NOTIFICATION = 3

DATA_SIZE = Qt.UserRole + 1


class SessionItem(QTreeWidgetItem):
    def __init__(self, l_list, is_session=False):
        QTreeWidgetItem.__init__(self, l_list)
        self.is_session = is_session
        self.setTextAlignment(3, Qt.AlignRight | Qt.AlignVCenter)

    def __lt__(self, other: 'SessionItem'):
        if self.childCount() and not other.childCount():
            return True

        if other.childCount() and not self.childCount():
            return False

        if OpenSessionDialog.sort_by_date:
            self_date_int = self.data(COLUMN_DATE, Qt.UserRole)
            other_date_int = other.data(COLUMN_DATE, Qt.UserRole)
            if self_date_int is None:
                if other_date_int is None:
                    return bool(self.text(COLUMN_NAME).lower()
                                < other.text(COLUMN_NAME).lower())
                return False
            
            if other_date_int is None:
                return True
            
            return self_date_int > other_date_int

        return bool(self.text(COLUMN_NAME).lower()
                    < other.text(COLUMN_NAME).lower())

    def show_conditionnaly(self, string: str) -> bool:
        show = bool(
            string.lower() in self.data(COLUMN_NAME, Qt.UserRole).lower())

        n = 0
        for i in range(self.childCount()):
            if self.child(i).show_conditionnaly(string.lower()):
                n += 1
        if n:
            show = True

        self.setExpanded(bool(n and string))
        self.setHidden(not show)
        return show

    def find_item_with(self, string: str) -> 'Optional[SessionItem]':
        if self.data(COLUMN_NAME, Qt.UserRole) == string:
            return self

        item = None

        for i in range(self.childCount()):
            item = self.child(i).find_item_with(string)
            if item:
                break

        return item
    
    def set_notes_icon(self, icon):
        self.setIcon(COLUMN_NOTES, icon)

    def set_scripted(self, script_flags:int, for_child=False):
        if script_flags == ray.ScriptFile.PREVENT:
            self.setText(COLUMN_SCRIPTS, "")
        else:
            if for_child:
                self.setText(COLUMN_SCRIPTS, "^_")
            else:
                self.setText(COLUMN_SCRIPTS, ">_")

        for i in range(self.childCount()):
            item = self.child(i)
            item.set_scripted(script_flags, for_child=True)
    
    def set_modified_date(self, date_int:int):
        self.setData(COLUMN_DATE, Qt.UserRole, date_int)

        date = QDateTime.fromSecsSinceEpoch(date_int)
        date_string =  date.toString("dd/MM/yy hh:mm")
        if QLocale.system().country() == QLocale.UnitedStates:
            date_string = date.toString("MM/dd/yy hh:mm")

        self.setText(COLUMN_DATE, date_string)
            
    def add_modified_date(self, path:str, date_int:int):
        self_path = self.data(COLUMN_NAME, Qt.UserRole)
        if not self_path:
            return
        
        if path == self_path:
            self.set_modified_date(date_int)
            return
        
        if path.startswith(self_path + '/'):
            current_data_date = self.data(COLUMN_DATE, Qt.UserRole)
            if current_data_date is None:
                current_data_date = 0
            
            if date_int > current_data_date:
                self.set_modified_date(date_int)
                
            for i in range(self.childCount()):
                child_item = self.child(i)
                child_item.add_modified_date(path, date_int)

    def set_locked(self, locked:bool):
        if locked:
            self.setFlags(self.flags() & ~Qt.ItemIsEnabled)
        else:
            self.setFlags(self.flags() | Qt.ItemIsEnabled)

    # Re-implemented protected Qt function
    def child(self, index: int) -> 'SessionItem':
        return QTreeWidgetItem.child(self, index)


class SessionFolder:
    name = ""
    path = ""
    is_session = False

    def __init__(self, name):
        self.name = name
        self.subfolders = list[SessionFolder]()
        self.item = None

    def set_path(self, path):
        self.path = path

    def make_item(self):
        self.item = SessionItem([self.name, "", "", ""], self.is_session)
        self.item.setData(COLUMN_NAME, Qt.UserRole, self.path)

        if self.subfolders:
            self.item.setIcon(COLUMN_NAME, QIcon.fromTheme('folder'))

        if not self.is_session:
            self.item.setFlags(self.item.flags() & ~Qt.ItemIsSelectable)

        for folder in self.subfolders:
            sub_item = folder.make_item()
            self.item.addChild(sub_item)

        return self.item

    def sort_childrens(self):
        if self.item is None:
            return
        
        self.item.sortChildren(COLUMN_NAME, Qt.AscendingOrder)
        
        for folder in self.subfolders:
            folder.sort_childrens()
            
    def find_item_with(self, sess_name:str):
        if self.item is None:
            return None
        
        return self.item.find_item_with(sess_name)


class SaveSessionTemplateDialog(child_dialogs.SaveTemplateSessionDialog):
    def __init__(self, parent):
        child_dialogs.SaveTemplateSessionDialog.__init__(self, parent)
        self._server_will_accept = True

    def _server_status_changed(self, server_status):
        # server will always accept, whatever the status
        pass
    
    def set_original_session_name(self, session_name:str):
        self.ui.labelLabel.setText(session_name)


class DuplicateDialog(child_dialogs.NewSessionDialog):
    def __init__(self, parent):
        child_dialogs.NewSessionDialog.__init__(
            self, parent, duplicate_window=True)
        self._server_will_accept = True
        self.ui.toolButtonFolder.setEnabled(False)
        self.ui.toolButtonFolder.setVisible(False)
        self._original_session_name = ''
    
    def _server_status_changed(self, server_status):
        # server will always accept, whatever the status
        pass
    
    def _add_sessions_to_list(self, session_names: list):
        child_dialogs.NewSessionDialog._add_sessions_to_list(self, session_names)
        if not session_names:
            subfolder, sep, after = self._original_session_name.rpartition('/')
            
            self.ui.lineEdit.setText(subfolder + sep)
    
    def set_original_session_name(self, session_name:str):
        self._original_session_name = session_name
        self.ui.labelOriginalSessionName.setText(session_name)


class OpenSessionDialog(ChildDialog):
    sort_by_date = False
    
    @classmethod
    def set_sort_by_date(cls, sort_by_date:bool):
        cls.sort_by_date = sort_by_date
    
    def __init__(self, parent):
        ChildDialog.__init__(self, parent)
        self.ui = ui.open_session.Ui_DialogOpenSession()
        self.ui.setupUi(self)

        self._pending_action = PENDING_ACTION_NONE
        self._session_renaming = ('', '')
        self._session_duplicating = ('', '')
        self._session_templating = ('', '')

        self._listing_timer_progress_n = 0
        self._listing_timer_progress = QTimer()
        self._listing_timer_progress.setInterval(50)
        self._listing_timer_progress.timeout.connect(self._listing_timer_progress_timeout)
        self._listing_timer_progress.start()
        self._progress_inverted = False

        self.session_menu = QMenu()
        self.action_duplicate = self.session_menu.addAction(
            QIcon.fromTheme('duplicate'),
            _translate('session_menu', 'Duplicate session'))
        self.action_save_as_template = self.session_menu.addAction(
            QIcon.fromTheme('template'),
            _translate('session_menu', 'Save session as template'))
        self.action_rename = self.session_menu.addAction(
            QIcon.fromTheme('edit-rename'),
            _translate('session_menu', 'Rename session'))
        self.action_remove = self.session_menu.addAction(
            QIcon.fromTheme('remove'),
            _translate('session_menu', 'Remove session'))

        dark = is_dark_theme(self)
        self.action_duplicate.setIcon(
            RayIcon('xml-node-duplicate', dark))
        self.action_save_as_template.setIcon(
            RayIcon('document-save-as-template', dark))

        self.action_rename.triggered.connect(self._ask_for_session_rename)
        self.action_duplicate.triggered.connect(self._ask_for_session_duplicate)
        self.action_save_as_template.triggered.connect(
            self._ask_for_session_save_as_template)
        self.action_remove.triggered.connect(self._ask_for_session_remove)
        self.ui.toolButtonSessionMenu.setMenu(self.session_menu)
        self.ui.toolButtonFolderPreview.clicked.connect(
            self._open_preview_folder)
        
        self.ui.splitterMain.setSizes([240, 800])
        self.ui.stackedWidgetSessionName.set_text('')
        self.ui.previewFrame.setEnabled(False)
        self.ui.tabWidget.tabBar().setExpanding(True)
        self.ui.toolButtonFolder.clicked.connect(self._change_root_folder)
        
        self.ui.splitterMain.splitterMoved.connect(
            self._splitter_moved)
        self.ui.stackedWidgetSessionName.name_changed.connect(
            self._session_name_changed)
        self.ui.sessionList.currentItemChanged.connect(
            self._current_item_changed)
        self.ui.sessionList.setFocus(Qt.OtherFocusReason)
        self.ui.sessionList.itemDoubleClicked.connect(self._go_if_any)
        self.ui.sessionList.itemClicked.connect(self._deploy_item)
        self.ui.sessionList.customContextMenuRequested.connect(
            self._show_context_menu)
        self.ui.filterBar.textEdited.connect(self._update_filtered_list)
        self.ui.filterBar.key_event.connect(self._up_down_pressed)
        self.ui.listWidgetPreview.properties_request.connect(
            self._show_client_properties)
        self.ui.listWidgetPreview.add_to_session_request.connect(
            self._add_client_to_current_session)
        self.ui.buttonBox.button(QDialogButtonBox.Ok).setEnabled(False)
        self.ui.currentSessionsFolder.setText(CommandLineArgs.session_root)
        self.ui.checkBoxShowDates.stateChanged.connect(
            self._set_full_sessions_view)
        self.ui.pushButtonCancelProgress.clicked.connect(
            self._cancel_copy_clicked)

        self.signaler.add_sessions_to_list.connect(self._add_sessions)
        self.signaler.root_changed.connect(self._root_changed)
        self.signaler.session_preview_update.connect(
            self._session_preview_update)
        self.signaler.session_details.connect(
            self._update_session_details)
        self.signaler.scripted_dir.connect(
            self._scripted_dir)
        self.signaler.parrallel_copy_state.connect(
            self._parrallel_copy_state)
        self.signaler.parrallel_copy_progress.connect(
            self._parrallel_copy_progress)
        self.signaler.parrallel_copy_aborted.connect(
            self._parrallel_copy_aborted)
        self.signaler.other_session_renamed.connect(
            self._session_renamed_by_server)
        self.signaler.other_session_duplicated.connect(
            self._session_duplicated_by_server)
        self.signaler.other_session_templated.connect(
            self._session_templated_by_server)

        self.to_daemon('/ray/server/list_sessions', 0)
        
        self.ui.groupBoxProgress.setVisible(False)

        if not self.daemon_manager.is_local:
            self.ui.toolButtonFolder.setVisible(False)
            self.ui.currentSessionsFolder.setVisible(False)
            self.ui.labelSessionsFolder.setVisible(False)

        self._server_will_accept = False
        self._has_selection = False
        self._last_mouse_click = 0
        self._last_session_item = None

        self._server_status_changed(self.session.server_status)

        self._set_preview_scripted(False)

        self.folders = list[SessionFolder]()
        self.all_items = []

        self.ui.filterBar.setFocus(Qt.OtherFocusReason)
        
        # snapshots related
        self.main_snap_group = SnapGroup()

        self._full_view = True        
        self._set_full_sessions_view(False)
        
        self._current_parrallel_copy_id = 0
        self._corner_mode = CORNER_HIDDEN
        self._set_corner_group(CORNER_HIDDEN)
        
        self.ui.checkBoxSaveCurrentSession.setVisible(
            self.session.server_status == ray.ServerStatus.READY)
        self.ui.listWidgetPreview.server_status_changed(
            self.session.server_status)
        
        self._last_selected_session = ''
        self._listing_sessions = False
    
    def _set_full_sessions_view(self, full_view:bool):
        self.ui.sessionList.setHeaderHidden(not full_view)
        
        if full_view:
            self.ui.sessionList.setColumnCount(4)
        else:
            self.ui.sessionList.setColumnCount(3)
            
        self._full_view = full_view
        self._resize_session_names_column()
        
        OpenSessionDialog.set_sort_by_date(full_view)
        root_item = self.ui.sessionList.invisibleRootItem()
        root_item.sortChildren(COLUMN_NAME, Qt.AscendingOrder)
        for folder in self.folders:
            folder.sort_childrens()
        
    def _server_status_changed(self, server_status):
        self.ui.toolButtonFolder.setEnabled(
            bool(server_status in (ray.ServerStatus.OFF,
                                   ray.ServerStatus.READY,
                                   ray.ServerStatus.CLOSE)))

        self._server_will_accept = bool(
            server_status in (
                ray.ServerStatus.OFF,
                ray.ServerStatus.READY) and not self.server_copying)

        if server_status != ray.ServerStatus.OFF:
            if self._root_folder_file_dialog is not None:
                self._root_folder_file_dialog.reject()
            self._root_folder_message_box.reject()

        self.ui.checkBoxSaveCurrentSession.setVisible(
            server_status == ray.ServerStatus.READY)
        
        self.ui.listWidgetPreview.server_status_changed(server_status)

        self._prevent_ok()

    def _listing_timer_progress_timeout(self):
        # server is listing sessions
        if self._listing_timer_progress_n >= 10: # 10 x 50ms = 500 ms
            # display groupBoxProgress only if listing takes at least 500ms
            # to prevent flircks
            self._set_corner_group(CORNER_LISTING)
        
        self.ui.progressBar.setValue(self._listing_timer_progress_n)
        if self._listing_timer_progress_n >= 100:
            self._listing_timer_progress_n = 0
            self._progress_inverted = not self._progress_inverted
            self.ui.progressBar.setInvertedAppearance(
                self._progress_inverted)
        self._listing_timer_progress_n += 5

    def _root_changed(self, session_root):
        self.ui.currentSessionsFolder.setText(session_root)
        self.ui.sessionList.clear()
        self.folders.clear()
        self.to_daemon('/ray/server/list_sessions', 0)

    def _add_sessions(self, session_names: list[str], out_of_listing=False):
        if not self._listing_sessions and not out_of_listing:
            # in case session server is listing sessions
            # but they are already listed.
            # Check which one is selected and clear all of them. 
            item = self.ui.sessionList.currentItem()
            if item is not None:
                self._last_selected_session = item.data(COLUMN_NAME, Qt.UserRole)
            
            self.folders.clear()
            self.ui.sessionList.clear()
            
        if not session_names:
            # there are no session_names here if session listing
            # is finished.
            self._listing_sessions = False
            self._listing_timer_progress.stop()
            height = self.ui.groupBoxProgress.size().height()
            self._set_corner_group(CORNER_HIDDEN)

            root_item = self.ui.sessionList.invisibleRootItem()
            sess_item = None
            
            if self._last_selected_session:
                for folder in self.folders:
                    item = folder.find_item_with(self._last_selected_session)
                    if item is not None:
                        self.ui.sessionList.setCurrentItem(item)
                        self.ui.sessionList.scrollToItem(item)
                        break
                else:
                    self._last_selected_session = ''
            
            if not self._last_selected_session:
                for sess in self.session.recent_sessions:
                    if sess == self.session.get_short_path():
                        continue

                    for i in range(root_item.childCount()):
                        item = root_item.child(i)
                        sess_item = item.find_item_with(sess)
                        if sess_item is not None:
                            break
                    
                    if sess_item is not None:
                        self.ui.sessionList.setCurrentItem(sess_item)
                        self.ui.sessionList.scrollToItem(sess_item)
                        break
                
            QTimer.singleShot(20, self._resize_session_names_column)
            return

        for session_name in session_names:
            folder_div = session_name.split('/')
            folders = self.folders

            for i in range(len(folder_div)):
                f = folder_div[i]
                for g in folders:
                    if g.name == f:
                        if i + 1 == len(folder_div):
                            g.set_path(session_name)
                            g.is_session = True

                        folders = g.subfolders
                        break
                else:
                    new_folder = SessionFolder(f)
                    if i + 1 == len(folder_div):
                        new_folder.set_path(session_name)
                        new_folder.is_session = True
                    else:
                        new_folder.set_path('/'.join(folder_div[:i+1]))

                    folders.append(new_folder)
                    folders = new_folder.subfolders

        if not out_of_listing:
            self._listing_sessions = True
        self.ui.sessionList.clear()

        for folder in self.folders:
            item = folder.make_item()
            self.ui.sessionList.addTopLevelItem(item)

        self.ui.sessionList.sortByColumn(COLUMN_NAME, Qt.AscendingOrder)

    def _update_filtered_list(self, filt):
        filter_text = self.ui.filterBar.displayText()
        root_item = self.ui.sessionList.invisibleRootItem()

        ## hide all non matching items
        for i in range(root_item.childCount()):
            sess_item: SessionItem = root_item.child(i)
            sess_item.show_conditionnaly(filter_text)

        # if selected item not in list, then select the first visible
        if (not self.ui.sessionList.currentItem()
                or self.ui.sessionList.currentItem().isHidden()):
            for i in range(root_item.childCount()):
                item = root_item.child(i)
                if not item.isHidden():
                    self.ui.sessionList.setCurrentItem(item)
                    break

        if (not self.ui.sessionList.currentItem()
                or self.ui.sessionList.currentItem().isHidden()):
            self.ui.filterBar.setStyleSheet(
                "QLineEdit { background-color: red}")
            self.ui.sessionList.setCurrentItem(None)
        else:
            self.ui.filterBar.setStyleSheet("")
            self.ui.sessionList.scrollTo(self.ui.sessionList.currentIndex())

    def _up_down_pressed(self, event):
        start_item = self.ui.sessionList.currentItem()
        QTreeWidget.keyPressEvent(self.ui.sessionList, event)
        if not start_item:
            return

        current_item = self.ui.sessionList.currentItem()
        if current_item == start_item:
            return

        ex_item = current_item

        while not current_item.flags() & Qt.ItemIsSelectable:
            ex_item = current_item
            QTreeWidget.keyPressEvent(self.ui.sessionList, event)
            current_item = self.ui.sessionList.currentItem()
            if current_item == ex_item:
                self.ui.sessionList.setCurrentItem(start_item)
                return

    def _current_item_changed(self, item: SessionItem, previous_item: SessionItem):
        self._has_selection = bool(item and item.data(COLUMN_NAME, Qt.UserRole))
        
        self.ui.listWidgetPreview.clear()
        self.ui.treeWidgetSnapshots.clear()
        self.ui.labelSessionSize.setText('')
        
        if item is not None and item.is_session:
            session_full_name = item.data(COLUMN_NAME, Qt.UserRole)
            self.ui.stackedWidgetSessionName.set_text(basename(session_full_name))
            self.ui.previewFrame.setEnabled(True)
            if session_full_name:
                self.to_daemon('/ray/server/get_session_preview', session_full_name)

            if item.text(COLUMN_SCRIPTS):
                self._set_preview_scripted(True)
            else:
                self._set_preview_scripted(False)
        else:
            self.ui.stackedWidgetSessionName.set_text('')
            self.ui.previewFrame.setEnabled(False)
            self._set_preview_scripted(False)

        if item is not None:
            self._last_current_item = item

        self._prevent_ok()

    def _set_preview_scripted(self, scripted:bool):
        if scripted:
            self.ui.labelPreviewScript.setText('>_')
            self.ui.labelPreviewScript.setToolTip(
                _translate('open_session', 'This session is scripted'))
            self.ui.labelPreviewScript.setStyleSheet(
                'QLabel{color:green;background-color:black}')
        else:
            self.ui.labelPreviewScript.setText('')
            self.ui.labelPreviewScript.setToolTip('')
            self.ui.labelPreviewScript.setStyleSheet('')

    def _prevent_ok(self):
        self.ui.buttonBox.button(QDialogButtonBox.Ok).setEnabled(
            bool(self._server_will_accept and self._has_selection))

    def _show_context_menu(self):
        item: SessionItem = self.ui.sessionList.currentItem()
        if item is None:
            return

        if not item.is_session:
            return

        x = QCursor.pos().x()
        rect = self.ui.sessionList.visualItemRect(item)
        y = self.ui.sessionList.mapToGlobal(rect.bottomLeft()).y()
        
        self.session_menu.exec(QPoint(x, y+1))

    def _set_pending_action(self, action:int):
        self._pending_action = action
        self._update_session_menu()

    def _open_preview_folder(self):
        item = self.ui.sessionList.currentItem()
        if item is None:
            return
        
        session_name = item.data(COLUMN_NAME, Qt.UserRole)
        self.to_daemon('/ray/server/open_file_manager_at',
                       os.path.join(CommandLineArgs.session_root, session_name))

    def _set_corner_group(self, corner_mode:int):
        self._corner_mode = corner_mode
        self.ui.groupBoxProgress.setVisible(corner_mode != CORNER_HIDDEN)
        self.ui.pushButtonCancelProgress.setVisible(corner_mode != CORNER_LISTING)

        if corner_mode == CORNER_LISTING:
            self.ui.labelProgress.setText(
                _translate('open_session', 'Listing sessions'))
        elif corner_mode == CORNER_COPY:
            self.ui.pushButtonCancelProgress.setText(
                _translate('open_session', 'Cancel'))
            self.ui.labelProgress.setText(
                _translate('open_session', 'Session copy'))
            self.ui.progressBar.setValue(0)
        elif corner_mode == CORNER_NOTIFICATION:
            self.ui.pushButtonCancelProgress.setText(
                _translate('open_session', 'Ok'))
            self.ui.labelProgress.setText(
                _translate('open_session', 'Session saved as template'))
            self.ui.progressBar.setValue(100)

    def _ask_for_session_rename(self):
        if not self.ui.previewFrame.isEnabled():
            return

        self.ui.stackedWidgetSessionName.toggle_edit()

    def _ask_for_session_duplicate(self):
        item = self.ui.sessionList.currentItem()
        if item is None:
            return
        
        old_session_name = item.data(COLUMN_NAME, Qt.UserRole)
        
        if self._pending_action:
            return

        dialog = DuplicateDialog(self)
        dialog.set_original_session_name(old_session_name)
        dialog.exec()
        if not dialog.result():
            return
        
        new_session_name = dialog.get_session_short_path()

        self._set_pending_action(PENDING_ACTION_DUPLICATE)
        self._session_duplicating = (old_session_name, new_session_name)
        self.to_daemon('/ray/session/duplicate_only', old_session_name,
                       new_session_name, CommandLineArgs.session_root)

    def _ask_for_session_save_as_template(self):
        item = self.ui.sessionList.currentItem()
        if item is None:
            return

        if self._pending_action:
            return

        session_name = item.data(COLUMN_NAME, Qt.UserRole)

        dialog = SaveSessionTemplateDialog(self)
        dialog.set_original_session_name(session_name)
        dialog.exec()
        if not dialog.result():
            return
        
        template_name = dialog.get_template_name()

        self._set_pending_action(PENDING_ACTION_TEMPLATE)
        self._session_templating = (session_name, template_name)
        self.to_daemon('/ray/server/save_session_template',
                       session_name, template_name)

    def _ask_for_session_remove(self):
        # we won't call the server to remove a session
        # because it would enable a very dangerous OSC path.
        # we will remove it directly in the GUI process and thread
        if CommandLineArgs.out_daemon:
            return

        # do not allow to remove from GUI a too big session
        # totally arbitrary choice : 95.37 Mb
        if self.session.preview_size >= 100000000:
            return
        
        item = self.ui.sessionList.currentItem()
        if item is None:
            return
        
        session_name = item.data(COLUMN_NAME, Qt.UserRole)
        full_path = os.path.join(CommandLineArgs.session_root, session_name)

        if not os.path.isdir(full_path):
            return
        
        ret = QMessageBox.critical(
            self,
            _translate('open_session', 'Remove session'),
            _translate('open_session',
                       '<p>Are you really sure to want to remove the  following session:</p>'
                       + '<p><strong>%s</strong></p>'
                       + '<p>This action is irreversible.')
            % session_name,
            QMessageBox.No | QMessageBox.Yes,
            QMessageBox.No)
        
        if ret != QMessageBox.Yes:
            return
            
        try:
            shutil.rmtree(full_path)
        except:
            # TODO
            return

        parent = item.parent()
        if parent is None:
            parent = self.ui.sessionList.invisibleRootItem()
        parent.removeChild(item)

    def _session_name_changed(self, new_name:str):
        item = self.ui.sessionList.currentItem()
        if item is None:
            return

        old_name = item.data(COLUMN_NAME, Qt.UserRole)

        # prevent accidental renaming to same name
        if basename(old_name) == new_name:
            return
        
        if self._pending_action:
            return

        self._set_pending_action(PENDING_ACTION_RENAME)
        self._session_renaming = (old_name, new_name)
        self.to_daemon('/ray/server/rename_session', old_name, new_name)

    def _parrallel_copy_state(self, session_id:int, state:int):
        if state and self._current_parrallel_copy_id:
            return
        
        self._current_parrallel_copy_id = session_id if state else 0

        if not state:
            self._set_corner_group(CORNER_HIDDEN)

    def _parrallel_copy_progress(self, session_id:int, progress:float):
        if session_id != self._current_parrallel_copy_id:
            return

        self._set_corner_group(CORNER_COPY)
        self.ui.progressBar.setValue(int(progress * 100))

    def _parrallel_copy_aborted(self):
        self._set_corner_group(CORNER_HIDDEN)
        self._set_pending_action(PENDING_ACTION_NONE)

    def _cancel_copy_clicked(self):
        if not self._current_parrallel_copy_id:
            if self._corner_mode == CORNER_NOTIFICATION:
                self._set_corner_group(CORNER_HIDDEN)
            return
        
        self.to_daemon('/ray/server/abort_parrallel_copy',
                       self._current_parrallel_copy_id)

    def _session_renamed_by_server(self):
        old_name, new_name = self._session_renaming
        self._session_renaming = ('', '')
        current_name = ''

        item = self.ui.sessionList.currentItem()
        if item is not None:
            current_name = item.data(COLUMN_NAME, Qt.UserRole)

        new_long_name = new_name
        if '/' in old_name:
            new_long_name = old_name.rpartition('/')[0] + '/' + new_name

        if current_name != old_name:
            # should rarely happens because rename session is very fast
            # in case session has been renamed but is not selected anymore
            for i in range(self.ui.sessionList.topLevelItemCount()):
                item: SessionItem = self.ui.sessionList.topLevelItem(i)
                session_item = item.find_item_with(old_name)
                if session_item is not None:
                    session_item.setData(COLUMN_NAME, Qt.UserRole, new_long_name)
                    session_item.setText(COLUMN_DATE, new_name)
                    break
            return
        
        if item is None:
            return

        item.setData(COLUMN_NAME, Qt.UserRole, new_long_name)
        item.setText(COLUMN_NAME, new_name)
        self.ui.stackedWidgetSessionName.set_text(new_name)
        self._set_pending_action(PENDING_ACTION_NONE)

    def _session_duplicated_by_server(self):
        old_name, new_name = self._session_duplicating
        self._session_duplicating = ('', '')
        self._set_pending_action(PENDING_ACTION_NONE)

        self._add_sessions([new_name], out_of_listing=True)

        for folder in self.folders:
            item = folder.find_item_with(new_name)
            if item is not None:
                self.ui.sessionList.setCurrentItem(item)
                filter_text = self.ui.filterBar.text()

                if filter_text.lower() in new_name.lower():
                    self._update_filtered_list('')
                else:
                    self.ui.filterBar.setText('')

                parent_item = item.parent()
                while parent_item is not None:
                    parent_item.setExpanded(True)
                    parent_item = parent_item.parent()
                self.ui.sessionList.scrollToItem(item)
                break

        self._set_corner_group(CORNER_HIDDEN)
    
    def _session_templated_by_server(self):
        session_name, template_name = self._session_templating
        self._session_templating = ('', '')
        if self._pending_action == PENDING_ACTION_TEMPLATE:
            self._set_corner_group(CORNER_NOTIFICATION)
        self._set_pending_action(PENDING_ACTION_NONE)
    
    def _deploy_item(self, item: SessionItem, column):
        if column == COLUMN_NOTES and not item.icon(COLUMN_NOTES).isNull():
            # set preview tab to 'Notes' tab if user clicked on a notes icon 
            self.ui.tabWidget.setCurrentIndex(1)

        if not item.childCount():
            return

        if time.time() - self._last_mouse_click > 0.35:
            item.setExpanded(not item.isExpanded())

        self._last_mouse_click = time.time()

    def _go_if_any(self, item: SessionItem, column):
        if item.childCount():
            return

        if (self._server_will_accept and self._has_selection
                and self.ui.sessionList.currentItem().data(
                    COLUMN_NAME, Qt.UserRole)):
            self.accept()

    def _session_preview_update(self):
        self.ui.plainTextEditNotes.setPlainText(self.session.preview_notes)

        for pv_client in self.session.preview_client_list:
            client_slot = self.ui.listWidgetPreview.create_client_widget(pv_client)
            client_slot.set_launched(
                pv_client.client_id in self.session.preview_started_clients)
        
        self.main_snap_group.snapshots.clear()
        self._add_snapshots(self.session.preview_snapshots)
        
        locale = QLocale()
        self.ui.labelSessionSize.setText(
            locale.formattedDataSize(self.session.preview_size))
        
        # store size in item
        item = self.ui.sessionList.currentItem()
        if item is not None:
            item.setData(COLUMN_NAME, DATA_SIZE, self.session.preview_size)
            self._set_preview_scripted(
                bool(item.text(COLUMN_SCRIPTS)))
        else:
            self._set_preview_scripted(False)
                
        self._update_session_menu()

    def _update_session_menu(self):
        item = self.ui.sessionList.currentItem()
        if item is None:
            self.session_menu.setEnabled(False)
            return
        
        self.session_menu.setEnabled(True)
        session_size = item.data(COLUMN_NAME, DATA_SIZE)
        allow_remove = False
        remove_title = _translate('session_menu', 'Remove session')
        
        if session_size is not None:
            if session_size >= 100000000:
                remove_title = _translate('session_menu', 'Remove session (too big)')
            else:
                allow_remove = True
        
        self.action_remove.setText(remove_title)
        
        ok = not bool(self._pending_action)
        self.action_duplicate.setEnabled(ok)
        self.action_rename.setEnabled(ok)
        self.action_save_as_template.setEnabled(ok)
        self.action_remove.setEnabled(ok and allow_remove)

    def _add_snapshots(self, snaptexts: list[str]):
        if not snaptexts and not self.main_snap_group.snapshots:
            # Snapshot list finished without any snapshot
            #self._no_snapshot_found()
            return

        for snaptext in snaptexts:
            if not snaptext:
                continue

            snapshot = Snapshot.new_from_snaptext(snaptext)
            self.main_snap_group.add(snapshot)

        self.main_snap_group.sort()

        self.ui.treeWidgetSnapshots.clear()

        for snapshot in self.main_snap_group.snapshots:
            item = snapshot.make_item(GROUP_MAIN)
            self.ui.treeWidgetSnapshots.addTopLevelItem(item)

        #self.ui.buttonBox.button(QDialogButtonBox.Ok).setEnabled(False)
        self.ui.treeWidgetSnapshots.clearSelection()

    def _update_session_details(self, session_name: str,
                                has_notes: int, modified: int, locked: int):
        for i in range(self.ui.sessionList.topLevelItemCount()):
            item: SessionItem = self.ui.sessionList.topLevelItem(i)
            session_item = item.find_item_with(session_name)
            if session_item is not None:
                if has_notes:
                    session_item.set_notes_icon(
                        RayIcon('notes', is_dark_theme(self)))
                
                # we add directly date to top item
                # this way folder also read the last date
                item.add_modified_date(session_name, modified)

                session_item.set_locked(bool(locked))
                break

    def _scripted_dir(self, dir_name, script_flags):
        if dir_name == '':
            # means that all the session root directory is scripted
            for i in range(self.ui.sessionList.topLevelItemCount()):
                item: SessionItem = self.ui.sessionList.topLevelItem(i)
                item.set_scripted(script_flags)
            return

        for i in range(self.ui.sessionList.topLevelItemCount()):
            item = self.ui.sessionList.topLevelItem(i)
            scripted_item = item.find_item_with(dir_name)
            if scripted_item is not None:
                scripted_item.set_scripted(script_flags)

    def _resize_session_names_column(self):
        self.ui.sessionList.setColumnWidth(COLUMN_NOTES, 20)
        self.ui.sessionList.setColumnWidth(COLUMN_SCRIPTS, 20)
        scroll_bar = self.ui.sessionList.verticalScrollBar()
        width = 20
        
        if self._full_view:
            self.ui.sessionList.setColumnWidth(COLUMN_DATE, 105)

            width = self.ui.sessionList.width() - 150
            if scroll_bar.isVisible():
                width -= scroll_bar.width()
            
            width = max(width, 20)
            
        else:
            width = self.ui.sessionList.width() - 45            
            if scroll_bar.isVisible():
                width -= scroll_bar.width()
            
            width = max(width, 40)

        self.ui.sessionList.setColumnWidth(COLUMN_NAME, width)

    def _show_client_properties(self, client_id:str):
        for pv_client in self.session.preview_client_list:
            if pv_client.client_id == client_id:
                properties_dialog = ClientPropertiesDialog.create(
                    self, pv_client)
                properties_dialog.update_contents()
                properties_dialog.lock_widgets()
                properties_dialog.show()
                break

    def _add_client_to_current_session(self, client_id:str):
        item = self.ui.sessionList.currentItem()
        if item is None:
            return
        
        session_name = item.data(COLUMN_NAME, Qt.UserRole)
        self.to_daemon('/ray/session/add_other_session_client',
                       session_name, client_id)
        self.reject()

    def _splitter_moved(self, pos:int, index:int):
        self._resize_session_names_column()

    def resizeEvent(self, event):
        ChildDialog.resizeEvent(self, event)
        self._resize_session_names_column()

    def closeEvent(self, event):
        self._cancel_copy_clicked()
        ChildDialog.closeEvent(self, event)

    def get_selected_session(self)->str:
        if self.ui.sessionList.currentItem():
            return self.ui.sessionList.currentItem().data(
                COLUMN_NAME, Qt.UserRole)

    def want_to_save_previous(self)->bool:
        if self.ui.checkBoxSaveCurrentSession.isHidden():
            return True

        return self.ui.checkBoxSaveCurrentSession.isChecked()
